# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements realistic constraints to be used with SinglePeriodOptimization
and MultiPeriodOptimization policies, or other Cvxpy-based policies.
"""


import cvxpy as cvx
import numpy as np

# from .utils import values_in_time
from .estimator import CvxpyExpressionEstimator, ParameterEstimator


__all__ = [
    "LongOnly",
    "LeverageLimit",
    "LongCash",
    "DollarNeutral",
    "ParticipationRateLimit",
    "MaxWeights",
    "MinWeights",
    "FactorMaxLimit",
    "FactorMinLimit",
    "FixedFactorLoading",
]


class BaseConstraint(CvxpyExpressionEstimator):
    """Base cvxpy constraint class."""
    
    #INITIALIZED = False # used to interface w/ old cvxportfolio

    # interface to old cvxportfolio
    def weight_expr(self, t, w_plus, z, v):
        #if not self.INITIALIZED:
        self.pre_evaluation(None, None, t, None)
        self.legacy_expression = self.compile_to_cvxpy(w_plus, z, v)
        #self.INITIALIZED = True
        self.values_in_time(t, None, None, None, None)
        if hasattr(self.legacy_expression, "__iter__"):
            return self.legacy_expression
        else:
            return [self.legacy_expression]


class BaseTradeConstraint(BaseConstraint):
    """Base class for constraints that operate on trades."""

    pass


class BaseWeightConstraint(BaseConstraint):
    """Base class for constraints that operate on weights.

    Here we can implement a method to pass benchmark weights
    and make the constraint relative to it rather than to the null
    portfolio.
    """

    pass


class ParticipationRateLimit(BaseTradeConstraint):
    """A limit on maximum trades size as a fraction of market volumes.

    Attributes:
        self.volumes (ParameterEstimator): ParameterEstimator with point-in-time market volumes estimations
        self.max_fraction_of_volumes (ParameterEstimator): ParameterEstimator with point-in-time,
             and also possibly per-stock, requirements of maximum participation rate

    """

    def __init__(self, volumes, max_fraction_of_volumes=0.05):
        self.volumes = ParameterEstimator(volumes)
        self.max_participation_rate = ParameterEstimator(
            max_fraction_of_volumes)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return cvx.multiply(cvx.abs(z[:-1]), portfolio_value) <= cvx.multiply(
            self.volumes, self.max_participation_rate
        )


class LongOnly(BaseWeightConstraint):
    """A long only constraint."""

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] >= 0


class LeverageLimit(BaseWeightConstraint):
    """A limit on leverage.

    Attributes:
      limit: A (time) series or scalar giving the leverage limit.
    """

    def __init__(self, limit):
        self.limit = ParameterEstimator(limit)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return cvx.norm(w_plus[:-1], 1) <= self.limit


class LongCash(BaseWeightConstraint):
    """Requires that cash be non-negative."""

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return w_plus[-1] >= 0


class DollarNeutral(BaseWeightConstraint):
    """Long-short dollar neutral strategy."""

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return w_plus[-1] == 1


class MaxWeights(BaseWeightConstraint):
    """A max limit on weights.

    Attributes:
      limit: A series or number giving the weights limit.
    """

    def __init__(self, limit):
        self.limit = ParameterEstimator(limit)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] <= self.limit


class MinWeights(BaseWeightConstraint):
    """A min limit on weights.

    Attributes:
      limit: A series or number giving the weights limit.
    """

    def __init__(self, limit):
        self.limit = ParameterEstimator(limit)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] >= self.limit


class FactorMaxLimit(BaseWeightConstraint):
    """A max limit on portfolio-wide factor (e.g. beta) exposure.

    Args:
        factor_exposure: An (n * r) matrix giving the factor exposure per asset
        per factor, where n represents # of assets and r represents # of factors
        limit: A series of list or a single list giving the factor limits
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = ParameterEstimator(factor_exposure)
        self.limit = ParameterEstimator(limit)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return self.factor_exposure.T @ w_plus[:-1] <= self.limit


class FactorMinLimit(BaseWeightConstraint):
    """A min limit on portfolio-wide factor (e.g. beta) exposure.

    Args:
        factor_exposure: An (n * r) matrix giving the factor exposure per asset
        per factor, where n represents # of assets and r represents # of factors
        limit: A series of list or a single list giving the factor limits
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = ParameterEstimator(factor_exposure)
        self.limit = ParameterEstimator(limit)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return self.factor_exposure.T @ w_plus[:-1] >= self.limit


class FixedFactorLoading(BaseWeightConstraint):
    """A constraint to fix portfolio loadings to a set of factors.

    This can be used to impose market neutrality, a certain portfolio-wide alpha, ....

    Attributes:
        factor_exposure: An (n * r) matrix giving the factor exposure on each
        factor
        target: A series or number giving the targeted factor loading
    """

    def __init__(self, factor_exposure, target):
        self.factor_exposure = ParameterEstimator(factor_exposure)
        self.target = ParameterEstimator(target)

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Return a Cvxpy constraint."""
        return self.factor_exposure.T @ w_plus[:-1] == self.target



# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains trading policies that can be backtested."""

import datetime as dt
import pandas as pd
import numpy as np
import logging
import cvxpy as cvx

from .costs import BaseCost
from .returns import BaseReturnsModel
from .constraints import BaseConstraint
#from cvxportfolio.utils import values_in_time, null_checker
from .estimator import Estimator, DataEstimator
from .errors import MissingValuesError, PortfolioOptimizationError
from .returns import ReturnsForecast

__all__ = [
    "Hold",
    "SellAll",
    "FixedTrades",
    "PeriodicRebalance",
    "ProportionalRebalance",
    "AdaptiveRebalance",
    "SinglePeriodOptimization",
    "MultiPeriodOptimization",
    "SinglePeriodOpt",
    "MultiPeriodOpt",
    "ProportionalTradeToTargets",
    "RankAndLongShort",
    "FixedWeights",
]


class BaseTradingPolicy(Estimator):
    """Base class for a trading policy."""

    # costs = []
    # constraints = []
    
    #INITIALIZED = False # used to interface w/ old cvxportfolio

    # TEMPORARY INTERFACE OLD NEW
    def get_trades(self, portfolio, t=dt.datetime.today()):
        """Trades list given current portfolio and time t."""
        value = sum(portfolio)
        w = pd.Series(portfolio, copy=True) / value
        # raise Exception
        #if not self.INITIALIZED:
        self.pre_evaluation(returns=pd.DataFrame(0.0, index=[t], columns=portfolio.index),
                volumes=None, start_time=t, end_time=None)
        #self.INITIALIZED = True
        return self.values_in_time(t, current_weights=w,
                current_portfolio_value=value, past_returns=None, past_volumes=None) * value

    def _nulltrade(self, portfolio):
        return pd.Series(index=portfolio.index, data=0.0)


class Hold(BaseTradingPolicy):
    """Hold initial portfolio, don't trade."""

    def values_in_time(self, t, current_weights, *args, **kwargs):
        """Update sub-estimators and produce current estimate."""
        return pd.Series(0.0, index=current_weights.index)


class RankAndLongShort(BaseTradingPolicy):
    """Rank assets by signal; long highest and short lowest.

    Args:
        signal (pd.DataFrame): time-indexed DataFrame of signal for all symbols
            excluding cash. At each point in time the num_long assets with
            highest signal will have equal positive weight, and the num_short assets with
            lower signal will have equal negative weight. If two or more assets have the same
            signal value and they are on the boundary of either the top or bottom set,
            alphanumerical ranking will prevail.
        num_long (int or pd.Series): number of assets to long, default 1; if specified as Series
            it must be indexed by time.
        num_short (int or pd.Series): number of assets to short, default 1; if specified as Series
            it must be indexed by time.
        target_leverage (float or pd.Series): leverage of the resulting portfolio, default 1;
            if specified as Series it must be indexed by time.
    """

    def __init__(self, signal, num_long=1, num_short=1, target_leverage=1.0):
        """Define sub-estimators at class attribute level."""
        self.num_long = DataEstimator(num_long)
        self.num_short = DataEstimator(num_short)
        self.signal = DataEstimator(signal)
        self.target_leverage = DataEstimator(target_leverage)

    def values_in_time(self, t, current_weights, *args, **kwargs):
        """Update sub-estimators and produce current estimate."""
        super().values_in_time(t, current_weights, *args, **kwargs)

        sorted_ret = pd.Series(
            self.signal.current_value, current_weights.index[:-1]
        ).sort_values()
        short_positions = sorted_ret.index[: self.num_short.current_value]
        long_positions = sorted_ret.index[-self.num_long.current_value:]

        target_weights = pd.Series(0.0, index=current_weights.index)
        target_weights[short_positions] = -1.0
        target_weights[long_positions] = 1.0

        target_weights /= sum(abs(target_weights))
        target_weights *= self.target_leverage.current_value

        # cash is always 1.
        target_weights[current_weights.index[-1]] = 1.0

        return target_weights - current_weights


class ProportionalTradeToTargets(BaseTradingPolicy):
    """Given a DataFrame of target weights in time, trade in equal proportions to reach those.

    Initially, it loads the list of trading days and so at each day it knows
    how many are missing before the next target's day, and trades in equal proportions
    to reach those targets. If there are no targets remaining it defaults to
    not trading.

    Args:
        targets (pd.DataFrame): time-indexed DataFrame of target weight vectors at
            given points in time (e.g., start of each month).

    """

    def __init__(self, targets):
        self.targets = targets

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Get list of trading days."""
        self.trading_days = returns.index
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)

    def values_in_time(self, t, current_weights, *args, **kwargs):
        """Get current trade weights."""
        super().values_in_time(t, current_weights, *args, **kwargs)
        next_targets = self.targets.loc[self.targets.index >= t]
        assert np.allclose(next_targets.sum(1), 1.)
        if not len(next_targets):
            return pd.Series(0.0, index=current_weights.index)
        next_target = next_targets.iloc[0]
        next_target_day = next_targets.index[0]
        trading_days_to_target = len(self.trading_days[(
            self.trading_days >= t) & (self.trading_days < next_target_day)])
        return (next_target - current_weights) / (trading_days_to_target + 1)


class SellAll(BaseTradingPolicy):
    """Sell all assets to cash.

    This is useful to check the tcost model in the simulator,
    or as an element in a (currently not implemented) composite policy.
    """

    def values_in_time(self, t, current_weights, *args, **kwargs):
        """Get current trade weights."""
        super().values_in_time(t, current_weights, *args, **kwargs)
        target = np.zeros(len(current_weights))
        target[-1] = 1.0
        return target - current_weights


class FixedTrades(BaseTradingPolicy):
    """Each day trade the provided trade weights vector.

    If there are no weights defined for the given day, default to no
    trades.

    Args:
        trades_weights (pd.Series or pd.DataFrame): Series of weights
            (if constant in time) or DataFrame of trade weights
            indexed by time. It trades each day the corresponding vector.
    """

    def __init__(self, trades_weights):
        """Trade the tradevec vector (dollars) or tradeweight weights."""
        self.trades_weights = DataEstimator(trades_weights)

    def values_in_time(self, t, current_weights, *args, **kwargs):
        try:
            super().values_in_time(t, current_weights, *args, **kwargs)
            return pd.Series(
                self.trades_weights.current_value,
                current_weights.index)
        except MissingValuesError:
            return pd.Series(0.0, current_weights.index)


class FixedWeights(BaseTradingPolicy):
    """Each day trade to the provided trade weights vector.

    If there are no weights defined for the given day, default to no
    trades.

    Args:
        target_weights (pd.Series or pd.DataFrame): Series of weights
            (if constant in time) or DataFrame of trade weights
            indexed by time. It trades each day to the corresponding vector.
    """

    def __init__(self, target_weights):
        """Trade the tradevec vector (dollars) or tradeweight weights."""
        self.target_weights = DataEstimator(target_weights)

    def values_in_time(self, t, current_weights, *args, **kwargs):
        try:
            super().values_in_time(t, current_weights, *args, **kwargs)
            return (
                pd.Series(
                    self.target_weights.current_value,
                    current_weights.index) -
                current_weights)
        except MissingValuesError:
            return pd.Series(0.0, current_weights.index)
            
class Uniform(FixedWeights):
    """Uniform allocation on non-cash assets."""
    
    def pre_evaluation(self, returns, *args, **kwargs):
        self.target_weights = pd.Series(1., returns.columns)
        self.target_weights.iloc[-1] = 0
        self.target_weights /= sum(self.target_weights)


class PeriodicRebalance(FixedWeights):
    """Track a target weight vector rebalancing at given times.

    This calls `FixedWeights`. If you want to change the target in time
    use that policy directly.


    Args:
        target (pd.Series): portfolio weights to rebalance to.
        rebalancing_times (pd.DateTimeIndex): after the open trading on these days
            portfolio is equal to target.
    """

    def __init__(self, target, rebalancing_times):
        target_weights = pd.DataFrame(
            {el: target for el in rebalancing_times}).T
        super().__init__(target_weights)


class ProportionalRebalance(ProportionalTradeToTargets):
    """Track a target weight exactly at given times, trading proportionally to it each period.

    This calls `ProportionalTradeToTargets`. If you want to change the target in time
    use that policy directly.

    Args:
        target (pd.Series): portfolio weights to rebalance to.
        target_matching_times (pd.DateTimeIndex): after the open trading on these days
            portfolio is equal to target.
    """

    def __init__(self, target, target_matching_times):
        targets = pd.DataFrame({el: target for el in target_matching_times}).T
        super().__init__(targets)


class AdaptiveRebalance(BaseTradingPolicy):
    """Rebalance portfolio when deviates too far from target.

    We use the 2-norm as trigger for rebalance. You may want to
    calibrate the `max_tracking_error` for your application
    by backtesting this policy, e.g., to get your desired turnover.

    Args:
        target (pd.Series or pd.DataFrame): target weights to rebalance to.
            It is assumed a constant if it is a Series. If it varies in
            time (you must specify it for every trading day) pass a
            DataFrame indexed by time.
        tracking_error (float or pd.Series): we trade to match the target
            weights whenever the 2-norm of our weights minus the
            target is larger than this. Pass a Series if you want to vary it in
            time.
    """

    def __init__(self, target, tracking_error):
        self.target = DataEstimator(target)
        self.tracking_error = DataEstimator(tracking_error)

    def values_in_time(self, t, current_weights, *args, **kwargs):
        super().values_in_time(t, current_weights, *args, **kwargs)
        if (
            np.linalg.norm(current_weights - self.target.current_value)
            > self.tracking_error.current_value
        ):
            return self.target.current_value - current_weights
        else:
            return pd.Series(0.0, current_weights.index)



class MultiPeriodOptimization(BaseTradingPolicy):
    """Multi Period Optimization policy.

    Implements the model developed in Chapter 5, in particular
    at page 49, of the book. You specify the objective terms
    using classes such as ReturnsForecast and TcostModel, each
    multiplied by its multiplier. You also specify lists
    of constraints. There are two ways to do it. You either
    define the same objective terms and costraints for each 
    step of the multi-period problem, or you define a different
    objective term and different list of constraints for each step.
    In addition we offer a `terminal_constraint` argument to
    simply impose that at the last step in the optimization the
    post-trade weights match the given weights (see page 51).
    
    When it computes the trajectory of weights for the future
    it only returns the first step (to the Simulator, typically).
    The future steps (planning horizon) are by default not returned.

    Args:
        objective (algebra of BaseCost or list of those): these will be maximized;
            if you pass a single expression of BaseCost it is understood as the 
            same for all steps; if it's a list you must also pass a list of lists
            for `constraints`, each term represents the cost for each step of the optimization
            (starting from the first, i.e., today) and the length of the list is 
            used as planning_horizon (the value you pass there will be ignored) 
        constraints (list of BaseConstraints or list of those): these will be
            imposed on the optimization. Default []. Pass this as a list of
            lists of the same length as `objective` to specify different 
            constraints at different time steps.
        planning_horizon (int or None): how many steps in the future we 
            plan for. Ignored if passing `objective` and `constraints` as lists.
            Default is None.
        terminal_constraint (pd.Series or None): if you pass a Series to this
            (default is None) it will impose that at the last step of the multi
            period optimization the post-trade weights are equal to this.
        **kwargs: these will be passed to cvxpy.Problem.solve,
            so you can choose your own solver and pass
            parameters to it.
    """

    def __init__(self, objective, constraints=[], planning_horizon=None,terminal_constraint=None,**kwargs):
        if hasattr(objective, '__iter__'):
            if not (hasattr(constraints, '__iter__') and len(constraints) and (hasattr(constraints[0], '__iter__') and len(objective) == len(constraints))):
                raise SyntaxError('If you pass objective as a list, constraints should be a list of lists of the same length.')
            self.planning_horizon = len(objective)
            self.objective = objective
            self.constraints = constraints
        else:
            if not np.isscalar(planning_horizon):
                raise SyntaxError('If `objective` and `constraints` are the same for all steps you must specify `planning_horizon`.')
            self.planning_horizon = planning_horizon
            self.objective = [objective for i in range(planning_horizon)]
            self.constraints = [constraints for i in range(planning_horizon)]
                
        self.terminal_constraint = terminal_constraint
        self.cvxpy_kwargs = kwargs

    def compile_to_cvxpy(self):#, w_plus, z, value):
        """Compile all cvxpy expressions and the problem."""
        self.cvxpy_objective = [
            el.compile_to_cvxpy(self.w_plus_at_lags[i], self.z_at_lags[i], self.portfolio_value) 
            for i, el in enumerate(self.objective)]
        self.cvxpy_objective = sum(self.cvxpy_objective)
        assert self.cvxpy_objective.is_dcp()  # dpp=True)
        assert self.cvxpy_objective.is_concave()
        self.cvxpy_constraints = [
            [constr.compile_to_cvxpy(self.w_plus_at_lags[i], self.z_at_lags[i], self.portfolio_value) 
                for constr in el]
            for i, el in enumerate(self.constraints)]
        self.cvxpy_constraints = sum(self.cvxpy_constraints, [])
        self.cvxpy_constraints += [cvx.sum(z) == 0 for z in self.z_at_lags]
        w = self.w_current
        for i in range(self.planning_horizon):
            self.cvxpy_constraints.append(self.w_plus_at_lags[i] == self.z_at_lags[i] + w)
            w = self.w_plus_at_lags[i]
        if not self.terminal_constraint is None:
            self.cvxpy_constraints.append(w == self.terminal_constraint)
        self.problem = cvx.Problem(cvx.Maximize(self.cvxpy_objective), self.cvxpy_constraints)
        assert self.problem.is_dcp()  # dpp=True)

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Pass a full view of the data to initialize objects that need it."""
        for obj in self.objective:
            obj.pre_evaluation(returns, volumes, start_time, end_time, **kwargs)
        for constr_at_lag in self.constraints:
            for constr in constr_at_lag:
                constr.pre_evaluation(returns, volumes, start_time, end_time, **kwargs)

        # initialize the problem
        self.portfolio_value = cvx.Parameter(nonneg=True)
        self.w_current = cvx.Parameter(returns.shape[1])
        self.z_at_lags = [cvx.Variable(returns.shape[1]) for i in range(self.planning_horizon)] 
        self.w_plus_at_lags = [cvx.Variable(returns.shape[1]) for i in range(self.planning_horizon)]

        self.compile_to_cvxpy()#self.w_plus, self.z, self.portfolio_value)

    def values_in_time(self, t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs):
        """Update all cvxpy parameters and solve."""
        
        assert current_portfolio_value > 0
        assert np.isclose(sum(current_weights), 1)
                
        for obj in self.objective:
            obj.values_in_time(t, current_weights, current_portfolio_value,
                past_returns, past_volumes, **kwargs)
        for constr_at_lag in self.constraints:
            for constr in constr_at_lag:
                constr.values_in_time(t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs)        

        self.portfolio_value.value = 1. # not used current_portfolio_value
        self.w_current.value = current_weights.values
        try:
            self.problem.solve(#ignore_dpp=True, 
                **self.cvxpy_kwargs)
        except cvx.SolverError:
            raise PortfolioOptimizationError(
                f"Numerical solver for policy {self.__class__.__name__} at time {t} failed;"
                "try changing it, relaxing some constraints, or dropping some costs.")
        if self.problem.status in ["unbounded", "unbounded_inaccurate"]:
            raise PortfolioOptimizationError(
                f"Policy {self.__class__.__name__} at time {t} resulted in an unbounded problem."
            )
        if self.problem.status in ["infeasible", 'infeasible_inaccurate']:
            raise PortfolioOptimizationError(
                f"Policy {self.__class__.__name__} at time {t} resulted in an infeasible problem."
            )

        return pd.Series(self.z_at_lags[0].value, current_weights.index)


class SinglePeriodOptimization(MultiPeriodOptimization):
    """Single Period Optimization policy.

    Implements the model developed in Chapter 4, in particular
    at page 43, of the book. You specify the objective term
    using classes such as ReturnsForecast and TcostModel, each
    multiplied by its multiplier. You also specify a list
    of constraints.

    Args:
        objective (algebra of BaseCost): this will be maximized.
        constraints (list of BaseConstraints): these will be
            imposed on the optimization. Default [].
        **kwargs: these will be passed to cvxpy.Problem.solve,
            so you can choose your own solver and pass
            parameters to it.
    """

    def __init__(self, objective, constraints=[], **kwargs):
        super().__init__([objective], [constraints], **kwargs)
        

class SinglePeriodOptNEW(SinglePeriodOptimization):
    """Placeholder class while we translate tests to new interface."""

    def __init__(
        self, return_forecast, costs, constraints, solver=None, solver_opts={}
    ):
        if np.isscalar(return_forecast):
            raise Exception
        if hasattr(return_forecast, "index"):
            return_forecast = ReturnsForecast(return_forecast)
        objective = -sum(costs, start=-return_forecast)
        kwargs = solver_opts
        if not (solver is None):
            kwargs["solver"] = solver
        super().__init__(objective, constraints, **kwargs)
        

class SinglePeriodOptOLD(BaseTradingPolicy):
    """Single-period optimization policy.

    Implements the model developed in chapter 4 of our paper
    https://stanford.edu/~boyd/papers/cvx_portfolio.html
    """

    def __init__(self, return_forecast,
            costs, constraints, solver=None, solver_opts=None):
        self.constraints = []
        self.costs = []
        if not isinstance(return_forecast, BaseReturnsModel):
            null_checker(return_forecast)
            self.return_forecast = ReturnsForecast(return_forecast)
        else:
            self.return_forecast = return_forecast

        super(SinglePeriodOpt, self).__init__()

        for cost in costs:
            assert isinstance(cost, BaseCost)
            self.costs.append(cost)

        for constraint in constraints:
            assert isinstance(constraint, BaseConstraint)                
            self.constraints.append(constraint)

        self.solver = solver
        self.solver_opts = {} if solver_opts is None else solver_opts

    def get_trades(self, portfolio, t=None):
        """
        Get optimal trade vector for given portfolio at time t.

        Parameters
        ----------
        portfolio : pd.Series
            Current portfolio vector.
        t : pd.timestamp
            Timestamp for the optimization.
        """

        if t is None:
            t = dt.datetime.today()

        value = sum(portfolio)
        w = portfolio / value
        z = cvx.Variable(w.size)  # TODO pass index
        wplus = w.values + z

        if isinstance(self.return_forecast, BaseReturnsModel):
            alpha_term = self.return_forecast.weight_expr(t, wplus, None, None)
        else:
            alpha_term = cvx.sum(
                cvx.multiply(
                    values_in_time(
                        self.return_forecast,
                        t).values,
                    wplus))

        assert alpha_term.is_concave()

        costs, constraints = [], []

        for cost in self.costs:
            cost_expr, const_expr = cost.weight_expr(t, wplus, z, value)
            costs.append(cost_expr)
            constraints += const_expr

        for constr in self.constraints:
            constraints += constr.weight_expr(t, wplus, z, value)

        for el in costs:
            assert el.is_convex()

        for el in constraints:
            assert el.is_dcp()

        # raise Exception

        self.prob = cvx.Problem(cvx.Maximize(
            alpha_term - sum(costs)), [cvx.sum(z) == 0] + constraints)
        try:
            self.prob.solve(solver=self.solver, ignore_dpp=True, **self.solver_opts)

            if self.prob.status in ["unbounded", "unbounded_inaccurate"]:
                logging.error(
                    "The problem is unbounded. Defaulting to no trades")
                return self._nulltrade(portfolio)

            if self.prob.status in ["infeasible", 'infeasible_inaccurate']:
                logging.error(
                    "The problem is infeasible. Defaulting to no trades")
                return self._nulltrade(portfolio)

            return pd.Series(index=portfolio.index, data=(z.value * value))
        except (cvx.SolverError, TypeError):
            logging.error(
                "The solver %s failed. Defaulting to no trades" %
                self.solver)
            return self._nulltrade(portfolio)


SinglePeriodOpt = SinglePeriodOptOLD


# class LookaheadModel():
#     """Returns the planning periods for multi-period.
#     """
#     def __init__(self, trading_times, period_lens):
#         self.trading_times = trading_times
#         self.period_lens = period_lens
#
#     def get_periods(self, t):
#         """Returns planning periods.
#         """
#         periods = []
#         tau = t
#         for length in self.period_lens:
#             incr = length*pd.Timedelta('1 days')
#             periods.append((tau, tau + incr))
#             tau += incr
#         return periods


class MultiPeriodOpt(SinglePeriodOpt):
    def __init__(
            self,
            trading_times,
            terminal_weights,
            lookahead_periods=None,
            *args,
            **kwargs):
        """
        trading_times: list, all times at which get_trades will be called
        lookahead_periods: int or None. if None uses all remaining periods
        """
        # Number of periods to look ahead.
        self.lookahead_periods = lookahead_periods
        self.trading_times = trading_times
        # Should there be a constraint that the final portfolio is the bmark?
        self.terminal_weights = terminal_weights
        super(MultiPeriodOpt, self).__init__(*args, **kwargs)

    def get_trades(self, portfolio, t=dt.datetime.today()):
        value = sum(portfolio)
        assert value > 0.0
        w = cvx.Constant(portfolio.values / value)

        prob_arr = []
        z_vars = []

        # planning_periods = self.lookahead_model.get_periods(t)
        for tau in self.trading_times[
            self.trading_times.get_loc(t): self.trading_times.get_loc(t)
            + self.lookahead_periods
        ]:
            # delta_t in [pd.Timedelta('%d days' % i) for i in
            # range(self.lookahead_periods)]:

            #            tau = t + delta_t
            z = cvx.Variable(*w.shape)
            wplus = w + z
            obj = self.return_forecast.weight_expr_ahead(t, tau, wplus)

            costs, constr = [], []
            for cost in self.costs:
                cost_expr, const_expr = cost.weight_expr_ahead(
                    t, tau, wplus, z, value)
                costs.append(cost_expr)
                constr += const_expr

            obj -= sum(costs)
            constr += [cvx.sum(z) == 0]
            for single_constr in self.constraints:
                constr += single_constr.weight_expr(t, wplus, z, value)

            prob = cvx.Problem(cvx.Maximize(obj), constr)
            prob_arr.append(prob)
            z_vars.append(z)
            w = wplus

        # Terminal constraint.
        if self.terminal_weights is not None:
            # prob_arr[-1].constraints += [wplus == self.terminal_weights.values]
            prob_arr[-1] = cvx.Problem(cvx.Maximize(obj),
                                       constr + [wplus == self.terminal_weights.values])

        sum(prob_arr).solve(solver=self.solver, ignore_dpp=True)
        return pd.Series(index=portfolio.index, data=(z_vars[0].value * value))



# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements cost functions used by optimization-based policies.

Currently these are two: TransactionCostModel and HoldingCostModel.

(In previous versions of Cvxportfolio these were used in the simulator as well,
but now instead we include their logic in the simulator itself.)

We do our best to include parameters that are accurate for the real market.
It should be easy to adjust these to other settings by following the description
provided in the class definitions.
"""

import cvxpy as cvx
import numpy as np
import pandas as pd
import copy

# from .expression import Expression
# from .utils import null_checker, values_in_time
from .estimator import CvxpyExpressionEstimator, ParameterEstimator

__all__ = ["HcostModel", "TcostModel"]


class BaseCost(CvxpyExpressionEstimator):
    """Base class for cost objects.

    It will use the CvxpyExpressionEstimator to compile the cost object to
    a cvxpy expression for optimization-based policies.

    It also overloads the values_in_time method to be used by simulator classes.
    """

    # gamma = 1. # this will be removed
    #LEGACY = False # used by some methods that need to know if they run in legacy mode
    #INITIALIZED = False # used to interface w/ old cvxportfolio
    
    # PLACEHOLDER METHOD TO USE OLD INTERFACE WITH NEW INTERFACE
    def weight_expr(self, t, w_plus, z, value):
        cost, constr = self._estimate(t, w_plus, z, value)
        return cost, constr

    def _estimate(self, t, w_plus, z, value):
        """Temporary interface to old cvxportfolio."""
        #self.LEGACY = True
        #if not self.INITIALIZED:
        placehoder_returns = pd.DataFrame(np.zeros((1, w_plus.shape[0] if not w_plus is None else z.shape[0])))
        self.pre_evaluation(placehoder_returns, None, t, None)
        self.legacy_expression = self.compile_to_cvxpy(w_plus, z, value)
        #self.INITIALIZED = True
        self.values_in_time(t, None, value, None, None)
        return self.legacy_expression, []

    def weight_expr_ahead(self, t, tau, w_plus, z, value):
        return self.weight_expr(t, w_plus, z, value)

    def __mul__(self, other):
        """Multiply by constant."""
        if not np.isscalar(other):
            raise SyntaxError("You can only multiply cost by a scalar.")
        return CombinedCosts([self], [other])

    def __add__(self, other):
        """Add cost expression to another cost expression.

        Idea is to create a new CombinedCost class that
        implements `compile_to_cvxpy` and values_in_time
        by summing over costs.

        """
        if isinstance(other, CombinedCosts):
            return other + self
        return CombinedCosts([self, other], [1.0, 1.0])

    def __rmul__(self, other):
        """Multiply by constant."""
        return self * other

    def __radd__(self, other):
        """Add cost expression to another cost."""
        return self + other

    def __neg__(self):
        """Take negative of cost."""
        return self * -1

    def __sub__(self, other):
        """Subtract other cost."""
        return self.__add__(-other)

    def __rsub__(self, other):
        """Subtract from other cost."""
        return other.__add__(-self)


class CombinedCosts(BaseCost):
    """Class obtained by algebraic combination of Cost classes.

    Attributes:
        costs (list): a list of BaseCost instances
    """

    def __init__(self, costs, multipliers):
        for cost in costs:
            if not isinstance(cost, BaseCost):
                raise SyntaxError(
                    "You can only sum `BaseCost` instances to other `BaseCost` instances."
                )
        self.costs = costs
        self.multipliers = multipliers
        
    # def _estimate(self, t, w_plus, z, value):
    #     """Temporary interface to old cvxportfolio."""
    #     #for cost in self.costs:
    #         #cost.LEGACY = True
    #     return super()._estimate(t, w_plus, z, value)

    def __add__(self, other):
        """Add other (combined) cost to self."""
        if isinstance(other, CombinedCosts):
            self.costs += other.costs
            self.multipliers += other.multipliers
        else:
            self.costs += [other]
            self.multipliers += [1.0]
        return self

    def __mul__(self, other):
        """Multiply by constant."""
        self.multipliers = [el * other for el in self.multipliers]
        return self

    def pre_evaluation(self, *args, **kwargs):
        """Iterate over constituent costs."""
        [el.pre_evaluation(*args, **kwargs) for el in self.costs]

    def values_in_time(self, *args, **kwargs):
        """Iterate over constituent costs."""
        [el.values_in_time(*args, **kwargs) for el in self.costs]

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Iterate over constituent costs."""
        self.expression = 0
        for multiplier, cost in zip(self.multipliers, self.costs):
            self.expression += multiplier * \
                cost.compile_to_cvxpy(w_plus, z, portfolio_value)
            # assert self.expression.is_dcp()#dpp=True)
        return self.expression
        # return sum([multiplier * cost.compile_to_cvxpy(w_plus, z, portfolio_value)
        #    for multiplier, cost in zip(self.multipliers, self.costs)])

    # TEMPORARY IN ORDER NOT TO BREAK TESTS
    # THESE METHODS ARE DEPRECATED

    def optimization_log(self, t):
        return sum(
            [
                multiplier
                * (cost.expression.value if hasattr(cost, "expression") else 0.0)
                for multiplier, cost in zip(self.multipliers, self.costs)
            ]
        )

    def simulation_log(self, t):
        return sum(
            [
                multiplier * (cost.last_cost if hasattr(cost, "last_cost") else 0.0)
                for multiplier, cost in zip(self.multipliers, self.costs)
            ]
        )


class HcostModel(BaseCost):
    """A model for holding costs.

    Attributes:
      borrow_costs: A dataframe of borrow costs.
      dividends: A dataframe of dividends.
    """

    def __init__(self, borrow_costs, dividends=0.0):
        self.borrow_costs = ParameterEstimator(borrow_costs, non_negative=True)
        self.dividends = ParameterEstimator(dividends)

    def compile_to_cvxpy(self, w_plus, z, value):
        """Compile cost to cvxpy expression."""
        self.expression = cvx.multiply(self.borrow_costs, cvx.neg(w_plus[:-1]))
        self.expression -= cvx.multiply(self.dividends, w_plus[:-1])
        self.expression = cvx.sum(self.expression)
        return self.expression

    def value_expr(self, t, h_plus, u):
        """Placeholder method as we update the rest of the stack to new interface."""
        
        #if not self.INITIALIZED:
        self.pre_evaluation(None, None, t, None)
        wplus = cvx.Variable(len(h_plus))
        z = cvx.Variable(len(h_plus))
        v = cvx.Parameter(nonneg=True)
        self.compile_to_cvxpy(wplus, z, v)
        self.values_in_time(t, None, None, None, None)

        self.last_cost = -np.minimum(0, h_plus.iloc[:-1]) * self.borrow_costs.value
        self.last_cost -= h_plus.iloc[:-1] * self.dividends.value

        return sum(self.last_cost)

    def optimization_log(self, t):
        return self.expression.value

    def simulation_log(self, t):
        return self.last_cost


class TcostModel(BaseCost):
    """A model for transaction costs.

    (See section pages 10-11 in the paper
    https://stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf).
    We don't include the short-term alpha term `c` here because it
    can be expressed with a separate `ReturnsForecast` object. If
    any term that appears in

    Args:
      half_spread (float or pd.Series or pd.DataFrame): Half the bid ask spread, either
        fixed per (non-cash) assets, or varying in time.
      nonlin_coeff (float or pd.Series or pd.DataFrame): Coefficients
            for the nonlinear cost term. This is the coefficient `b` in the paper.
            It can be constant, constant per-stock, or varying in time and stocks. Default 0.
      sigma (floar or pd.Series or pd.DataFrame): Daily volatilities. Default 0.
      volume (float or pd.Series or pd.DataFrame): Market volumes expressed in value (e.g., dollars).
            Default 1 to avoid NaN.
      power (float): The nonlinear tcost exponent. Default 1.5.
    """

    def __init__(self, half_spread=0.0, nonlin_coeff=0.0, sigma=0.0, volume=1.0, power=1.5):
        # self.compile_first_term = not np.isscalar(half_spread) or half_spread > 0.0
        # self.compile_second_term = (not np.isscalar(nonlin_coeff) or nonlin_coeff > 0.0) and (not np.isscalar(sigma) or sigma > 0.0)
        # if self.compile_first_term:
        self.half_spread = ParameterEstimator(half_spread, non_negative=True)
        # self.sigma = DataEstimator(sigma) #, non_negative=True)
        # self.volume = DataEstimator(volume) #, non_negative=True)
        # self.nonlin_coeff = DataEstimator(nonlin_coeff) #, non_negative=True)
        #if self.compile_second_term:
        self.base_second_term_multiplier = ParameterEstimator(sigma * nonlin_coeff / (volume**(power - 1)))#, non_negative=True)
        self.power: float = power
            
    def pre_evaluation(self, *args, **kwargs):
        super().pre_evaluation(*args, **kwargs)
        #if self.compile_second_term:
        self.second_term_multiplier = cvx.Parameter(shape=self.base_second_term_multiplier.shape, nonneg=True)

    def values_in_time(self, t, current_weights, current_portfolio_value, past_returns, 
        past_volumes, **kwargs):
        """We patch here to apply current portfolio value to tcost."""
        super().values_in_time(t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs)
        # self.second_term_multiplier.value *= current_portfolio_value ** (self.power - 1)
        #if self.compile_second_term:
        self.second_term_multiplier.value = self.base_second_term_multiplier.value * current_portfolio_value ** (self.power - 1)

    def compile_to_cvxpy(self, w_plus, z, value):
        #if self.compile_first_term:
        #first_term = (cvx.multiply(self.half_spread, cvx.abs(z[:-1])) if self.compile_first_term else 0.0)
        first_term = cvx.multiply(self.half_spread, cvx.abs(z[:-1]))
        assert cvx.sum(first_term).is_convex()
        #else:
        #    first_term = 0.
        #if self.compile_second_term:
            # second_term = cvx.multiply(self.nonlin_coeff, self.sigma)
            # second_term = cvx.multiply(
            #     second_term, (value / self.volume) ** (self.power - 1)
            # )
        second_term = cvx.multiply(self.second_term_multiplier, cvx.abs(z[:-1]) ** self.power)
        assert cvx.sum(second_term).is_convex()
        assert cvx.sum(second_term).is_dcp(dpp=True)
        #else:
        #    second_term = 0.0

        self.expression = first_term + second_term
        self.expression = cvx.sum(self.expression)
        assert self.expression.is_dcp(dpp=True)
        assert self.expression.is_convex()
        return self.expression

    # def _estimate(self, t, w_plus, z, value):
    #     """Estimate tcosts given trades.
    #
    #     Args:
    #       t: time of estimate
    #       z: trades
    #       value: portfolio value
    #
    #     Returns:
    #       An expression for the tcosts.
    #     """
    #
    #     z = z[:-1]
    #
    #     constr = []
    #
    #     second_term = (
    #         values_in_time(self.nonlin_coeff, t)
    #         * values_in_time(self.sigma, t)
    #         * (value / values_in_time(self.volume, t)) ** (self.power - 1)
    #     )
    #
    #     # # no trade conditions
    #     # if np.isscalar(second_term):
    #     #     if np.isnan(second_term):
    #     #         constr += [z == 0]
    #     #         second_term = 0
    #     # else:  # it is a pd series
    #     #     no_trade = second_term.index[second_term.isnull()]
    #     #     second_term[no_trade] = 0
    #     #     constr += [z[second_term.index.get_loc(tick)] == 0 for tick in no_trade]
    #
    #     try:
    #         self.expression = cvx.multiply(
    #             values_in_time(self.half_spread, t), cvx.abs(z)
    #         )
    #     except TypeError:
    #         self.expression = cvx.multiply(
    #             values_in_time(self.half_spread, t).values, cvx.abs(z)
    #         )
    #     try:
    #         self.expression += cvx.multiply(second_term, cvx.abs(z) ** self.power)
    #     except TypeError:
    #         self.expression += cvx.multiply(
    #             second_term.values, cvx.abs(z) ** self.power
    #         )
    #
    #     return cvx.sum(self.expression), constr

    # THESE METHODS ARE DEPRECATED AND WILL BE REMOVED AS WE FINISH
    # TRANSLATING TO NEW INTERFACE
    def value_expr(self, t, h_plus, u):
        """Temporary placeholder, new simulators implement their own tcost."""

        #if not self.INITIALIZED:
        self.pre_evaluation(None, None, t, None)
        wplus = cvx.Variable(len(u))
        z = cvx.Variable(len(u))
        v = cvx.Parameter(nonneg=True)
        self.compile_to_cvxpy(wplus, z, v)
        self.values_in_time(t, None, 1., None, None)

        u_nc = u.iloc[:-1]
        self.tmp_tcosts = 0.
        #if self.compile_first_term:
        self.tmp_tcosts += np.abs(u_nc) * self.half_spread.value
        #if self.compile_second_term:
        self.tmp_tcosts += (
                self.second_term_multiplier.value
                # self.nonlin_coeff.value
                # * self.sigma.value
                * np.abs(u_nc) ** self.power
                # / (self.volume.value ** (self.power - 1))
            )

        return self.tmp_tcosts.sum()

    def optimization_log(self, t):
        try:
            return self.expression.value
        except AttributeError:
            return np.nan

    def simulation_log(self, t):
        # TODO find another way
        return self.tmp_tcosts

    def est_period(self, t, tau_start, tau_end, w_plus, z, value):
        """Returns the estimate at time t of tcost over given period."""
        K = (tau_end - tau_start).days
        tcost, constr = self.weight_expr(t, None, z / K, value)
        return tcost * K, constr



# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains classes that define return models for
portfolio optimization policies, and related objects.
"""

import cvxpy as cvx
import numpy as np

# from cvxportfolio.expression import Expression
# from .legacy import values_in_time, null_checker
from .costs import BaseCost, CombinedCosts
from .risks import BaseRiskModel
from .estimator import DataEstimator, ParameterEstimator

__all__ = [
    "ReturnsForecast",
    "MultipleReturnsForecasts",
    "RollingWindowReturnsForecast",
    "ExponentialWindowReturnsForecast",
    "ReturnsForecastErrorRisk",
    "RollingWindowReturnsForecastErrorRisk",
]


class BaseReturnsModel(BaseCost):
    """Base class for return models.

    Use this to define any logic common to return models.
    """

    # interface to old
    def weight_expr(self, t, w_plus, z=None, value=None):
        cost, constr = self._estimate(t, w_plus, z, value)
        return cost


class ReturnsForecast(BaseReturnsModel):
    """Simple return forecast provided by the user.

    Args:
        expected_returns (pd.DataFrame or pd.Series): constant per-symbol
            forecasts (if Series) or varying in time (if DataFrame)
    """

    def __init__(self, expected_returns, name=None):
        self.expected_returns = ParameterEstimator(expected_returns)
        self.name = name

    def compile_to_cvxpy(self, w_plus, z, v):
        return w_plus.T @ self.expected_returns
        
        
class MultipleReturnsForecasts(BaseReturnsModel):
    """A weighted combination of alpha sources.

    DEPRECATED: THIS CLASS IS KEPT TO REUSE OLD TESTS
    IT'S NOW IMPLEMENTED BY COMBINED COSTS, I.E.,
    YOU JUST WRITE AN ALGEBRA OF RETURNS FORECASTS
    """

    def __init__(self, alpha_sources, weights):
        self.alpha_sources = alpha_sources
        self.weights = weights
    
    def pre_evaluation(self, *args, **kwargs):
        """Iterate over constituent costs."""
        [el.pre_evaluation(*args, **kwargs) for el in self.alpha_sources]

    def values_in_time(self, *args, **kwargs):
        """Iterate over constituent costs."""
        [el.values_in_time(*args, **kwargs) for el in self.alpha_sources] 
               
    def compile_to_cvxpy(self, w_plus, z, v):
        return sum([el.compile_to_cvxpy(w_plus, z, v) * self.weights[i] for i, el in enumerate(self.alpha_sources)])
        
    def _estimate(self, t, w_plus, z, value):
        """Temporary interface to old cvxportfolio."""
        #for cost in self.alpha_sources:
        #    cost.LEGACY = True
        return super()._estimate(t, w_plus, z, value)

    # def weight_expr(self, t, wplus, z=None, v=None):
    #     """Returns the estimated alpha.
    #
    #     Args:
    #         t: time estimate is made.
    #         wplus: An expression for holdings.
    #         tau: time of alpha being estimated.
    #
    #     Returns:
    #       An expression for the alpha.
    #     """
    #     alpha = 0
    #     for idx, source in enumerate(self.alpha_sources):
    #         alpha += source.weight_expr(t, wplus) * self.weights[idx]
    #     return alpha
    #
    # def weight_expr_ahead(self, t, tau, wplus):
    #     """Returns the estimate at time t of alpha at time tau.
    #
    #     Args:
    #       t: time estimate is made.
    #       wplus: An expression for holdings.
    #       tau: time of alpha being estimated.
    #
    #     Returns:
    #       An expression for the alpha.
    #     """
    #     alpha = 0
    #     for idx, source in enumerate(self.alpha_sources):
    #         alpha += source.weight_expr_ahead(t,
    #                                           tau, wplus) * self.weights[idx]
    #     return alpha



class RollingWindowReturnsForecast(ReturnsForecast):
    """Compute returns forecast by rolling window mean of past returns.

    Args:
        lookback_period (int): how many past returns are used at each point in time.
            Default is 250.
        use_last_for_cash (bool): for the cash return instead just use the last
            value. Default True.
    """

    def __init__(self, lookback_period, use_last_for_cash=True):
        self.lookback_period = lookback_period
        self.use_last_for_cash = use_last_for_cash

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        forecasts = returns.rolling(
            window=self.lookback_period).mean().shift(1)
        if self.use_last_for_cash:
            forecasts.iloc[:, -1] = returns.iloc[:, -1].shift(1)
        self.expected_returns = ParameterEstimator(forecasts)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class ExponentialWindowReturnsForecast(ReturnsForecast):
    """Compute returns forecast by exponential window mean of past returns.

    Args:
        half_life (int): Half life of exponential decay used. Default is 250.
        use_last_for_cash (bool): for the cash return instead just use the last
            value. Default True.
    """

    def __init__(self, half_life, use_last_for_cash=True):
        self.half_life = half_life
        self.use_last_for_cash = use_last_for_cash

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        forecasts = returns.ewm(halflife=self.half_life).mean().shift(1)
        if self.use_last_for_cash:
            forecasts.iloc[:, -1] = returns.iloc[:, -1].shift(1)
        self.expected_returns = ParameterEstimator(forecasts)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class ReturnsForecastErrorRisk(BaseRiskModel):
    """Simple return forecast error risk with values provided by the user.

    Implements the model described in pages 31-32 of the paper. You
    must pass the delta Series (if constant) or DataFrame (if time-varying)
    of the forecast errors. Or, use one of the classes below to
    compute it automatically. Multiply this object by an external multiplier
    (which can itself be a ParameterEstimator) and calibrate for the right
    amount of penalization.

    Args:
        deltas_errors (pd.DataFrame or pd.Series): constant per-symbol
            errors on the returns forecasts (if Series),
             or varying in time (if DataFrame).
    """

    def __init__(self, deltas_errors):
        self.deltas_errors = ParameterEstimator(deltas_errors)

    def compile_to_cvxpy(self, w_plus, z, v):
        return cvx.abs(w_plus - self.benchmark_weights).T @ self.deltas_errors


class RollingWindowReturnsForecastErrorRisk(ReturnsForecastErrorRisk):
    """Compute returns forecast errors with rolling window of past returns.

    We compute the forecast error as the standard deviation of the mean
    estimator on a rolling window. That is, the rolling window standard deviation
    divided by sqrt(lookback_period), i.e., the square root of the number of samples
    used.

    Args:
        lookback_period (int): how many past returns are used at each point in time.
            Default is 250.
        zero_for_cash (bool): for the cash return forecast instead the error is zero.
    """

    def __init__(self, lookback_period, zero_for_cash=True):
        self.lookback_period = lookback_period
        self.zero_for_cash = zero_for_cash

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        tmp = returns.rolling(window=self.lookback_period).std().shift(
            1) / np.sqrt(self.lookback_period)
        if self.zero_for_cash:
            tmp.iloc[:, -1] = 0.0
        self.deltas_errors = ParameterEstimator(tmp, non_negative=True)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class ExponentialWindowReturnsForecastErrorRisk(ReturnsForecastErrorRisk):
    """Compute returns forecast errors with exponential window of past returns.

    Currently not implemented; we need to work out the math. It's probably
    as simple as ewm(...).std() / np.sqrt(half_life).
    """

    def __init__(self):
        raise NotImplementedError





# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains classes that have been replaced by improved versions.
They are kept because they have some slight differences and are used by 
the original examples scripts to generate the book's plots.
"""

import matplotlib.pyplot as plt
import pandas as pd
import datetime as dt
import logging
import cvxpy as cvx
import numpy as np

from .policies import BaseTradingPolicy
from .returns import BaseReturnsModel, MultipleReturnsForecasts
from .costs import BaseCost
from .estimator import ParameterEstimator

logger = logging.getLogger(__name__)

__all__ = [
    "null_checker",
    "non_null_data_args",
    "values_in_time",
    "plot_what_if",
    "LegacyReturnsForecast",
    "MPOReturnsForecast",
    ]


def values_in_time(obj, t, tau=None):
    """Obtain value(s) of object at time t, or right before.

    Optionally specify time tau>=t for which we want a prediction,
    otherwise it is assumed tau = t.

    obj: callable, pd.Series, pd.DataFrame, or something else.

        If a callable, we return obj(t,tau).

        If obj has an index attribute,
        we try to return obj.loc[t],
        or obj.loc[t, tau], if the index is a MultiIndex.
        If not available, we return obj.

        Otherwise, we return obj.

    t: np.Timestamp (or similar). Time at which we want
        the value.

    tau: np.Timestamp (or similar), or None. Time tau >= t
        of the prediction,  e.g., tau could be tomorrow, t
        today, and we ask for prediction of market volume tomorrow,
        made today. If None, then it is assumed tau = t.

    """

    if hasattr(obj, "__call__"):
        return obj(t, tau)

    if isinstance(obj, pd.Series) or isinstance(obj, pd.DataFrame):
        try:
            if not (tau is None) and isinstance(obj.index, pd.MultiIndex):
                return obj.loc[(t, tau)]
            else:
                return obj.loc[t]
        except KeyError:
            return obj

    return obj


def plot_what_if(time, true_results, alt_results):
    true_results.value.plot(label=true_results.pol_name)
    for result in alt_results:
        result.value.plot(label=result.pol_name, linestyle="--")
    plt.axvline(x=time, linestyle=":")


def null_checker(obj):
    """Check if obj contains NaN."""
    if isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
        if np.any(pd.isnull(obj)):
            raise ValueError("Data object contains NaN values", obj)
    elif np.isscalar(obj):
        if np.isnan(obj):
            raise ValueError("Data object contains NaN values", obj)
    else:
        raise TypeError("Data object can only be scalar or Pandas.")


def non_null_data_args(f):
    def new_f(*args, **kwds):
        for el in args:
            null_checker(el)
        for el in kwds.values():
            null_checker(el)
        return f(*args, **kwds)

    return new_f


# LEGACY CLASSES USED BY OLD TESTS. WILL BE REMOVED AS WE FINISH TRANSLATION


class LegacyReturnsForecast(BaseReturnsModel):
    """A single return forecast.

    STILL USED BY OLD PARTS OF CVXPORTFOLIO

    Attributes:
      alpha_data: A dataframe of return estimates.
      delta_data: A confidence interval around the estimates.
      half_life: Number of days for alpha auto-correlation to halve.
    """

    def __init__(self, returns, delta=0.0, gamma_decay=None, name=None):
        #null_checker(returns)
        self.returns = ParameterEstimator(returns)
        #null_checker(delta)
        self.delta = ParameterEstimator(delta)
        self.gamma_decay = gamma_decay
        self.name = name
        
    def compile_to_cvxpy(self, wplus, z, value):
        alpha = cvx.multiply(self.returns, wplus)
        alpha -= cvx.multiply(self.delta, cvx.abs(wplus))
        return cvx.sum(alpha)

    # def weight_expr(self, t, wplus, z=None, v=None):
    #     """Returns the estimated alpha.
    #
    #     Args:
    #       t: time estimate is made.
    #       wplus: An expression for holdings.
    #       tau: time of alpha being estimated.
    #
    #     Returns:
    #       An expression for the alpha.
    #     """
    #     alpha = cvx.multiply(values_in_time(self.returns, t), wplus)
    #     alpha -= cvx.multiply(values_in_time(self.delta, t), cvx.abs(wplus))
    #     return cvx.sum(alpha)

    def weight_expr_ahead(self, t, tau, wplus):
        """Returns the estimate at time t of alpha at time tau.

        Args:
          t: time estimate is made.
          wplus: An expression for holdings.
          tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """

        alpha = self.weight_expr(t, wplus)
        if tau > t and self.gamma_decay is not None:
            alpha *= (tau - t).days ** (-self.gamma_decay)
        return alpha


class MPOReturnsForecast(BaseReturnsModel):
    """A single alpha estimation.

    Attributes:
      alpha_data: A dict of series of return estimates.
    """

    def __init__(self, alpha_data):
        self.alpha_data = alpha_data

    def weight_expr_ahead(self, t, tau, wplus):
        """Returns the estimate at time t of alpha at time tau.

        Args:
          t: time estimate is made.
          wplus: An expression for holdings.
          tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """
        return self.alpha_data[(t, tau)].values.T * wplus


# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements Estimator base classes. Policies, costs, and constraints inherit
from this.
"""

import numpy as np
import pandas as pd
from .errors import MissingValuesError, DataError
import cvxpy


class Estimator:
    """Estimator base class.

    Policies, costs, and constraints inherit from this. When overloading
    methods defined here one should be careful on deciding whether to call
    the `super()` corresponding method. It can make sense to call it before
    some logic, after, or not calling it at all. Also, any subclass of this that uses
    logic defined here should be careful to put estimator subclasses at the class
    attribute level, so that the two methods defined here get called recursively
    on them.
    """

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Run computation on estimators before the simulation with full prescience.

        This function is called by Simulator classes before the
        start of a backtest with the full dataset available to the
        simulator. This is useful for estimators such as pandas.rolling_mean
        which are faster (and easier) when vectorized rather than called separately
        at each point in time.

        It should be used very carefully, if you
        are not sure do not implement this method. You can use
        values_in_time to build lazily whatever you would
        build here beforehand. If you do implement this, double
        check that at each point in time only past data
        (with respect to that point) is used and not future data.

        Args:
            returns (pandas.DataFrame): market returns.
            volumes (pandas.DataFrame): market volumes.
            start_time (pandas.Timestamp): start time of the backtest.
            end_time (pandas.Timestamp): end time of the backtest.
            **kwargs: extra arguments for future development.
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "pre_evaluation"):
                subestimator.pre_evaluation(
                    returns, volumes, start_time, end_time, **kwargs)

    def values_in_time(
            self,
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs):
        """Evaluates estimator at a point in time recursively on its sub-estimators.

        This function is called by Simulator classes on Policy classes
        returning the current trades list. Policy classes, if
        they contain internal estimators, should declare them as attributes
        and call this base function (via `super()`) before
        they do their internal computation. CvxpyExpression estimators
        should instead define this method to update their Cvxpy parameters.

        Args:
            t (pd.TimeStamp): point in time of the simulation.
            current_weights (pd.Series): current portfolio weights.
            current_portfolio_value (float): current total value of the portfolio.
            past_returns (pd.DataFrame): view of the market returns up to today (i.e., the
                last row are equal to today's open prices divided by yesterday's, minus 1).
            past_volumes (pd.DataFrame): view of the market volumes up to yesterday's.
            kwargs (dict): extra arguments for future development.
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "values_in_time"):
                subestimator.values_in_time(
                    t,
                    current_weights,
                    current_portfolio_value,
                    past_returns,
                    past_volumes,
                    **kwargs)


class CvxpyExpressionEstimator(Estimator):
    """Base class for estimators that are Cvxpy expressions."""

    def compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Compile term to cvxpy expression.

        This is called by a Policy class on its terms before the start of the backtest
        to compile its Cvxpy problem. If the Policy changes in time
        this is called at every time step.

        It can either return a scalar expression, in the case of objective terms,
        or a list of cvxpy constraints, in the case of constraints.

        In MultiPeriodOptimization policies this is called separately
        for costs and constraints at different look-ahead steps with
        the corresponding w_plus and z.

        Args:
            w_plus (cvxpy.Variable): post-trade allocation weights vector
            z (cvxpy.Variable): trades weight vector
            portfolio_value (cvxpy.Parameter): scalar Parameter that holds the value of the portfolio

        Returns:
            cvxpy.Expression
        """
        raise NotImplementedError


class DataEstimator(Estimator):
    """Estimator of point-in-time values from internal `self.data`.

    It also implements logic to check that no `np.nan` are returned
    by its `values_in_time` method, which is the way `cvxportfolio`
    objects use this class to get data.

    Args:
        data (object, pandas.Series, pandas.DataFrame): Data expressed
            preferably as pandas Series or DataFrame where the first
            index is a pandas.DateTimeIndex. Otherwise you can
            pass a callable object which implements the values_in_time method
            (with the standard signature) and returns the corresponding value in time,
             or a constant float, numpy.array, or even pandas Series or DataFrame not
            indexed by time (e.g., a covariance matrix where both index and columns
            are the stock symbols).
        use_last_available_time (bool): if the pandas index exists
            and is a pandas.DateTimeIndex you can instruct self.values_in_time
            to retrieve the last available value at time t by setting
            this to True. Default is False.

    """

    def __init__(self, data, use_last_available_time=False):
        self.data = data
        self.use_last_available_time = use_last_available_time

    def value_checker(self, result):
        """Ensure that only scalars or arrays without np.nan are returned.

        Args:
            result (int, float, or np.array): data produced by self.values_in_time

        Returns:
            result (int, float, or np.array): same data if no np.nan are present and type is correct

        Raises:
            cvxportfolio.MissingValuesError: if np.nan's are present in result
            cvxportfolio.DataError: if data is not in the right form
        """

        if np.isscalar(result):
            if np.isnan(result):
                raise MissingValuesError(
                    f"{self.__class__.__name__}.values_in_time result is a np.nan scalar."
                )
            else:
                return result

        if isinstance(result, np.ndarray):
            if np.any(np.isnan(result)):
                raise MissingValuesError(
                    f"{self.__class__.__name__}.values_in_time result is an array with np.nan's."
                )
            else:
                return result

        raise DataError(
            f"{self.__class__.__name__}.values_in_time result is not a scalar or array."
        )

    def internal_values_in_time(self, t, *args, **kwargs):
        """Internal method called by `self.values_in_time`."""

        if hasattr(self.data, "values_in_time"):
            return self.value_checker(
                self.data.values_in_time(
                    t, *args, **kwargs))

        if (
            hasattr(self.data, "loc")
            and hasattr(self.data, "index")
            and (
                isinstance(self.data.index, pd.DatetimeIndex)
                or (
                    isinstance(self.data.index, pd.MultiIndex)
                    and isinstance(self.data.index.levels[0], pd.DatetimeIndex)
                )
            )
        ):
            try:
                if self.use_last_available_time:
                    if isinstance(self.data.index, pd.MultiIndex):
                        newt = self.data.index.levels[0][
                            self.data.index.levels[0] <= t
                        ][-1]
                    else:
                        newt = self.data.index[self.data.index <= t][-1]
                    tmp = self.data.loc[newt]
                else:
                    tmp = self.data.loc[t]
                if hasattr(tmp, "values"):
                    return self.value_checker(tmp.values)
                else:
                    return self.value_checker(tmp)

            except (KeyError, IndexError):
                raise MissingValuesError(
                    f"{self.__class__.__name__}.values_in_time could not find data for requested time."
                )

        if hasattr(self.data, "values"):
            return self.value_checker(self.data.values)

        return self.value_checker(self.data)

    def values_in_time(self, t, *args, **kwargs):
        """Obtain value of `self.data` at time t or right before.

        Args:
            t (pandas.TimeStamp): time at which we evaluate the estimator

        Returns:
            result (float, numpy.array): if you use a callable object make
                sure that it returns a float or numpy array (and not,
                for example, a pandas object)

        """
        self.current_value = self.internal_values_in_time(t, *args, **kwargs)
        return self.current_value


# class ConstantEstimator(cvxpy.Constant, DataEstimator):
#     """Cvxpy constant that uses the pre_evalution method to be initialized."""
#
#     def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
#         """You should call super().__init__ it here."""
#         raise NotImplementedError


class ParameterEstimator(cvxpy.Parameter, DataEstimator):
    """Data estimator of point-in-time values that contains a Cvxpy Parameter.

    Attributes:
        parameter (cvxpy.Parameter): the parameter object to use with cvxpy
            expressions
    Args:
        same as cvxportfolio.DataEstimator

    """

    def __init__(
        self,
        data,
        positive_semi_definite=False,
        non_negative=False,
        use_last_available_time=False,
    ):
        self.positive_semi_definite = positive_semi_definite
        self.non_negative = non_negative
        self.use_last_available_time = use_last_available_time
        self.data = data
        # super(DataEstimator).__init__(data, use_last_available_time)

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Use the start time of the simulation to initialize the Parameter."""
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)
        value = super().values_in_time(start_time, None, None, None, None)
        super().__init__(
            value.shape if hasattr(value, "shape") else (),
            PSD=self.positive_semi_definite,
            nonneg=self.non_negative,
        )
        # self.parameter = self
        # self.parameter = cvxpy.Parameter(value.shape if hasattr(value, 'shape') else (), PSD=self.positive_semi_definite, nonneg=self.non_negative)

    def values_in_time(self, t, *args, **kwargs):
        """Update Cvxpy Parameter value."""
        self.value = super().values_in_time(t, *args, **kwargs)



# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import collections
import numpy as np
import pandas as pd
import copy
from .legacy import *
from .policies import SinglePeriodOpt, MultiPeriodOpt
from .estimator import Estimator

__all__ = ['BackTest', 'SimulationResult']

class BacktestResult(Estimator):
    """This will be the class returned by the simulator.

    It defines a bunch of metrics (Sharpe, ...) that are properties
    computed from its attribytes. It uses the Estimator logic to update
    its values by running its registered Simulator class.

    Attributes:
        w (pd.DataFrame): pre-trade weights
        z (pd.DataFrame): trade weights
        v (pd.Series): value of the portfolio

    """

    pass


def getFiscalQuarter(dt):
    """Convert a time to a fiscal quarter."""
    year = dt.year
    quarter = (dt.month - 1) // 3 + 1
    return "Q%i %s" % (quarter, year)
    
    
class CommonComputations(BacktestResult):
    """Temporary while having both old and new interface."""

    @property
    def v(self):
        """The value of the portfolio over time."""
        return self.h.sum(axis=1)

    @property
    def profit(self):
        """The profit made, in dollars."""
        return self.v[-1] - self.v[0]

    @property
    def w(self):
        """The weights of the portfolio over time."""
        return (self.h.T / self.v).T

    @property
    def leverage(self):
        """Portfolio leverage"""
        return np.abs(self.w).sum(1)

    @property
    def volatility(self):
        """The annualized, realized portfolio volatility."""
        return np.sqrt(self.PPY) * np.std(self.returns)

    @property
    def mean_return(self):
        """The annualized mean portfolio return."""
        return self.PPY * np.mean(self.returns)

    @property
    def returns(self):
        """The returns R_t = (v_{t+1}-v_t)/v_t"""
        val = self.v
        return pd.Series(
            data=val.values[1:] / val.values[:-1] - 1, index=val.index[:-1]
        )

    @property
    def growth_rates(self):
        """The growth rate log(v_{t+1}/v_t)"""
        return np.log(self.returns + 1)

    @property
    def annual_growth_rate(self):
        """The annualized growth rate PPY/T sum_{t=1}^T log(v_{t+1}/v_t)"""
        return self.growth_rates.sum() * self.PPY / self.growth_rates.size

    @property
    def annual_return(self):
        """The annualized return in percent."""
        ret = self.growth_rates
        return self._growth_to_return(ret.mean())

    def _growth_to_return(self, growth):
        """Convert growth to annualized percentage return."""
        return 100 * (np.exp(self.PPY * growth) - 1)

    def get_quarterly_returns(self, benchmark=None):
        """The annualized returns for each fiscal quarter."""
        ret = self.growth_rates
        quarters = ret.groupby(getFiscalQuarter).aggregate(np.mean)
        return self._growth_to_return(quarters)

    def get_best_quarter(self, benchmark=None):
        ret = self.get_quarterly_returns(benchmark)
        return (ret.argmax(), ret.max())

    def get_worst_quarter(self, benchmark=None):
        ret = self.get_quarterly_returns(benchmark)
        return (ret.argmin(), ret.min())

    @property
    def sharpe_ratio(self):
        return (
            np.sqrt(self.PPY)
            * np.mean(self.excess_returns)
            / np.std(self.excess_returns)
        )

    @property
    def turnover(self):
        """Turnover ||u_t||_1/v_t"""
        noncash_trades = self.u.drop(self.cash_key, axis=1)
        return np.abs(noncash_trades).sum(axis=1) / self.v.loc[self.u.index]

    @property
    def trading_days(self):
        """The fraction of days with nonzero turnover."""
        return (self.turnover.values > 0).sum() / self.turnover.size

    @property
    def max_drawdown(self):
        """The maximum peak to trough drawdown in percent."""
        val_arr = self.v.values
        max_dd_so_far = 0
        cur_max = val_arr[0]
        for val in val_arr[1:]:
            if val >= cur_max:
                cur_max = val
            elif 100 * (cur_max - val) / cur_max > max_dd_so_far:
                max_dd_so_far = 100 * (cur_max - val) / cur_max
        return max_dd_so_far


class SimulationResult(CommonComputations):
    """A container for the result of a simulation.

    Attributes:
        h_next: A dataframe of holdings over time.
        u: A dataframe of trades over time.
        tcosts: A series of transaction costs over time.
        borrow_costs: A series of borrow costs over time.
    """

    def __init__(
        self,
        initial_portfolio,
        policy,
        cash_key,
        simulator,
        simulation_times=None,
        PPY=252,
        timedelta=pd.Timedelta("1 days"),
    ):
        """
        Initialize the result object.

        Args:
            initial_portfolio:
            policy:
            simulator:
            simulation_times:
            PPY:
        """
        self.PPY = PPY
        self.timedelta = timedelta
        self.initial_val = sum(initial_portfolio)
        self.initial_portfolio = copy.copy(initial_portfolio)
        self.cash_key = cash_key
        self.simulator = simulator
        self.policy = policy

    def summary(self):
        print(self._summary_string())

    def _summary_string(self):
        data = collections.OrderedDict(
            {
                "Number of periods": self.u.shape[0],
                "Initial timestamp": self.h.index[0],
                "Final timestamp": self.h.index[-1],
                "Portfolio return (%)": self.returns.mean() * 100 * self.PPY,
                "Excess return (%)": self.excess_returns.mean() * 100 * self.PPY,
                "Excess risk (%)": self.excess_returns.std() * 100 * np.sqrt(self.PPY),
                "Sharpe ratio": self.sharpe_ratio,
                "Max. drawdown": self.max_drawdown,
                "Turnover (%)": self.turnover.mean() * 100 * self.PPY,
                "Average policy time (sec)": self.policy_time.mean(),
                "Average simulator time (sec)": self.simulation_time.mean(),
            }
        )

        return pd.Series(data=data).to_string(float_format="{:,.3f}".format)

    def log_data(self, name, t, entry):
        try:
            getattr(self, name).loc[t] = entry
        except AttributeError:
            setattr(
                self,
                name,
                (pd.Series if np.isscalar(entry) else pd.DataFrame)(
                    index=[t], data=[entry]
                ),
            )

    def log_policy(self, t, exec_time):
        self.log_data("policy_time", t, exec_time)
        # TODO mpo policy requires changes in the optimization_log methods
        if not isinstance(self.policy, MultiPeriodOpt):
            for cost in self.policy.costs:
                self.log_data(
                    "policy_" +
                    cost.__class__.__name__,
                    t,
                    cost.optimization_log(t))

    def log_simulation(self, t, u, h_next, risk_free_return, exec_time):
        self.log_data("simulation_time", t, exec_time)
        self.log_data("u", t, u)
        self.log_data("h_next", t, h_next)
        self.log_data("risk_free_returns", t, risk_free_return)
        for cost in self.simulator.costs:
            self.log_data(
                "simulator_" +
                cost.__class__.__name__,
                t,
                cost.simulation_log(t))

    @property
    def h(self):
        """
        Concatenate initial portfolio and h_next dataframe.

        """
        tmp = self.h_next.copy()
        tmp.loc["last"] = np.nan
        tmp = self.h_next.shift(1)
        tmp.iloc[0] = self.initial_portfolio
        # TODO fix ?
        # tmp.loc[self.h_next.index[-1] + self.timedelta]=self.h_next.iloc[-1]
        return tmp

    
        
        
class BackTest(CommonComputations):
    """Perform a backtest and hold the results."""
    
    def __init__(self, policy, simulator, start_time, end_time=None, value_init = 1E6, h=None):
        
        self.policy = policy
        self.simulator = simulator
        self.start_time = pd.Series(simulator.returns.data.index >= start_time, simulator.returns.data.index).idxmax()
        if end_time is None:
            self.end_time  = simulator.returns.data.index[-1]
        else:
            self.end_time = simulator.returns.data.index[simulator.returns.data.index <= end_time][-1]
        
        self.end_time = end_time
        simulator.initialize_policy(policy, self.start_time , self.end_time)
        
        if h is None:
            h = pd.Series(0., simulator.returns.data.columns)
            h[-1] = value_init
        
        self.h = pd.DataFrame(columns=simulator.returns.data.columns)
        self.u = pd.DataFrame(columns=simulator.returns.data.columns)
        self.z = pd.DataFrame(columns=simulator.returns.data.columns)
        self.tcost = pd.Series(dtype=float)
        self.hcost_stocks = pd.Series(dtype=float)
        self.hcost_cash = pd.Series(dtype=float)
        
        for t in simulator.returns.data.index[(simulator.returns.data.index >= self.start_time) & (simulator.returns.data.index < self.end_time)]:
            self.h.loc[t] = h
            h, self.z.loc[t], self.u.loc[t], self.tcost.loc[t], self.hcost_stocks.loc[t], self.hcost_cash.loc[t] = \
                simulator.simulate(t=t, h=h, policy=self.policy)
        
        self.h.loc[self.end_time] = h    
        
        self.PPY = 252
        self.timedelta = pd.Timedelta('1d')
        self.cash_key = self.h.columns[-1]
        
    @property
    def excess_returns(self):
        return self.returns - self.simulator.returns.data[self.cash_key].loc[self.returns.index]

        



# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .estimator import ParameterEstimator, DataEstimator  # , ConstantEstimator
import logging

import scipy.linalg

import cvxpy as cvx
import numpy as np
import pandas as pd

from .costs import BaseCost
# from .utils import values_in_time

logger = logging.getLogger(__name__)


__all__ = [
    "FullCovariance",
    "RollingWindowFullCovariance",
    "ExponentialWindowFullCovariance",
    "DiagonalCovariance",
    "RollingWindowDiagonalCovariance",
    "ExponentialWindowDiagonalCovariance",
    "LowRankRollingRisk",
    "RollingWindowFactorModelRisk",
    "WorstCaseRisk",
    "FactorModelRisk",
    # "CashBenchMark",
]


# class CashBenchMark(ConstantEstimator):
#     """Default benchmark weights for cvxportfolio risk models.
#     """
#
#     def __init__(self):
#         pass
#
#     def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
#         """Initialize it using the size of the returns.
#         """
#         size = returns.shape[1]
#         value = np.zeros(size)
#         value[-1] = 1.
#         super().__init__(value)
#         super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class BaseRiskModel(BaseCost):
    # benchmark_weights = None

    # DEPRECATED,BENCHMARK WEIGHTS ARE NOW PASSED BY set_benchmark
    def __init__(self, **kwargs):
        self.w_bench = kwargs.pop("w_bench", 0.0)
        self.benchmark_weights = None #self.w_bench
        # super(BaseRiskModel, self).__init__()
        # self.gamma_half_life = kwargs.pop("gamma_half_life", np.inf)

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        if not hasattr(self, 'benchmark_weights') or self.benchmark_weights is None:
            bw = pd.Series(0.0, returns.columns)
            bw.iloc[-1] = 1.0
            self.benchmark_weights = bw  # ParameterEstimator(bw)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)

    def set_benchmark(self, benchmark_weights):
        """We can only have constant benchmark because otherwise it is not dpp compliant.

        DEPRECATED: IT SHOULD NOT BE PASSED HERE. IT SHOULD BE PASSED TO POLICY
            AND ADDED w_plus_wrt_bm as a variable with equality constraint.
        """
        self.benchmark_weights = benchmark_weights  # ParameterEstimator(benchmark_weights)

    def weight_expr(self, t, w_plus, z, value):
        """Temporary placeholder while migrating to new interface"""
        self.expression, _ = self._estimate(t, w_plus - self.w_bench, z, value)
        return self.expression, []

    # DEPRECATED, MAYBE INCLUDE ITS LOGIC AT BASECOST LEVEL
    def optimization_log(self, t):
        if self.expression.value:
            return self.expression.value
        else:
            return np.NaN


class FullCovariance(BaseRiskModel):
    """Quadratic risk model with full covariance matrix.

    Args:
        Sigma (pd.DataFrame): DataFrame of Sigma matrices as understood
        by `cvxportfolio.estimator.DataEstimator`.
        forecast_error_kappa (float or pd.Series): add cost term for the
            uncertainty on the assets' correlations. See the paper, pages 32-33.
            Default is zero, corresponding to no uncertainty cost.
    """

    def __init__(self, Sigma=None, forecast_error_kappa=0.0, **kwargs):
        # DEPRECATED, IT'S USED BY SOME OLD CVXPORTFOLIO PIECES
        super(FullCovariance, self).__init__(**kwargs)
        self.Sigma = ParameterEstimator(Sigma, positive_semi_definite=True)
        self.forecast_error_kappa = DataEstimator(
            forecast_error_kappa)  # ParameterEstimator(
        self.parameter_forecast_error = cvx.Parameter(
            Sigma.shape[1], nonneg=True)
        #    forecast_error_kappa, non_negative=True
        # )

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)
        self.Sigma_sqrt = cvx.Parameter(self.Sigma.shape)

    def values_in_time(self, t, current_weights, current_portfolio_value, past_returns, past_volumes,
            **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""
        super().values_in_time(t, current_weights, current_portfolio_value,
            past_returns, past_volumes, **kwargs)
        self.parameter_forecast_error.value = np.sqrt(
            np.diag(self.Sigma.value)) * np.sqrt(self.forecast_error_kappa.current_value)
        #if not self.LEGACY:
        self.Sigma_sqrt.value = scipy.linalg.sqrtm(self.Sigma.value)
        assert np.allclose(
            self.Sigma.value,
            self.Sigma_sqrt.value @ self.Sigma_sqrt.value.T)

    def compile_to_cvxpy(self, w_plus, z, value):
        # something's broken with the old interface, patching it here
        #if self.LEGACY:
        #self.cvxpy_expression = cvx.quad_form(w_plus - self.benchmark_weights, self.Sigma) 
        #else:
        self.cvxpy_expression =  cvx.sum_squares(self.Sigma_sqrt.T @ (w_plus - self.benchmark_weights))
        
        # assert self.cvxpy_expression.is_dcp(dpp=True)
        self.cvxpy_expression += cvx.square(cvx.abs(w_plus - self.benchmark_weights).T @ self.parameter_forecast_error)
        # assert self.cvxpy_expression.is_dcp(dpp=True)
        return self.cvxpy_expression


class RollingWindowFullCovariance(FullCovariance):
    """Build FullCovariance model automatically with pandas rolling window.

    At the start of a backtest receives a view of asset returns
    (excluding cash) and computes the rolling window covariance
    for given lookback_period (default 250).

    Args:
        loockback_period (int): how many past returns are used each
            trading day to compute the historical covariance.
    """

    def __init__(
            self,
            lookback_period=250,
            zero_cash_covariance=True,
            forecast_error_kappa=0.0):
        self.lookback_period = lookback_period
        self.zero_cash_covariance = zero_cash_covariance
        self.forecast_error_kappa = DataEstimator(forecast_error_kappa)

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Function to initialize object with full prescience."""
        # drop cash return
        if self.zero_cash_covariance:
            returns = returns.copy(deep=True)
            returns.iloc[:, -1] = 0.0
        self.Sigma = ParameterEstimator(
            # shift forward so only past returns are used
            returns.rolling(
                window=self.lookback_period).cov().shift(
                returns.shape[1]),
            positive_semi_definite=True,
        )
        # initialize cvxpy Parameter(s)
        self.parameter_forecast_error = cvx.Parameter(
            returns.shape[1], nonneg=True)

        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class ExponentialWindowFullCovariance(FullCovariance):
    """Build FullCovariance model automatically with pandas exponential window.

    At the start of a backtest receives a view of asset returns
    (excluding cash) and computes the exponential window covariance
    for given half_life (default 250).

    Args:
        half_life (int): the half life of exponential decay used by pandas
            exponential moving window
    """

    def __init__(
            self,
            half_life=250,
            zero_cash_covariance=True,
            forecast_error_kappa=0.0):
        self.half_life = half_life
        self.zero_cash_covariance = zero_cash_covariance
        self.forecast_error_kappa = DataEstimator(forecast_error_kappa)

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Function to initialize object with full prescience."""

        if self.zero_cash_covariance:
            returns = returns.copy(deep=True)
            returns.iloc[:, -1] = 0.0
        self.Sigma = ParameterEstimator(
            # shift forward so only past returns are used
            returns.ewm(halflife=self.half_life).cov().shift(returns.shape[1]),
            positive_semi_definite=True,
        )
        self.parameter_forecast_error = cvx.Parameter(
            returns.shape[1], nonneg=True)
        # initialize cvxpy Parameter(s)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class DiagonalCovariance(BaseRiskModel):
    """Risk model using diagonal covariance matrix.

    Args:
        standard_deviations (pd.DataFrame or pd.Series): per-stock standard
            deviations, defined as square roots of covariances.
            Indexed by time if DataFrame.
    """

    def __init__(self, standard_deviations=None):
        self.standard_deviations = ParameterEstimator(standard_deviations)

    def compile_to_cvxpy(self, w_plus, z, value):
        return cvx.sum_squares(
            cvx.multiply(
                w_plus -
                self.benchmark_weights,
                self.standard_deviations))


class RollingWindowDiagonalCovariance(DiagonalCovariance):
    """Build DiagonalCovariance model automatically with pandas rolling window.

    At the start of a backtest receives a view of asset returns
    (excluding cash) and computes the rolling window variances
    for given lookback_period (default 250).

    Args:
        loockback_period (int): how many past returns are used each
            trading day to compute the historical covariance.
    """

    def __init__(self, lookback_period=250, zero_cash_covariance=True):
        self.lookback_period = lookback_period
        self.zero_cash_covariance = zero_cash_covariance

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Function to initialize object with full prescience."""
        # drop cash return
        if self.zero_cash_covariance:
            returns = returns.copy(deep=True)
            returns.iloc[:, -1] = 0.0
        self.standard_deviations = ParameterEstimator(
            # shift forward so only past returns are used
            np.sqrt(
                returns.rolling(window=self.lookback_period).var().shift(1),
            )
        )
        # initialize cvxpy Parameter(s)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class ExponentialWindowDiagonalCovariance(DiagonalCovariance):
    """Build DiagonalCovariance model automatically with pandas exponential window.

    At the start of a backtest receives a view of asset returns
    (excluding cash) and computes the exponential window variances
    for given half_life (default 250).

    Args:
        half_life (int): the half life of exponential decay used by pandas
            exponential moving window
    """

    def __init__(self, half_life=250, zero_cash_covariance=True):
        self.half_life = half_life
        self.zero_cash_covariance = zero_cash_covariance

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Function to initialize object with full prescience."""

        # drop cash return
        if self.zero_cash_covariance:
            returns = returns.copy(deep=True)
            returns.iloc[:, -1] = 0.0
        self.standard_deviations = ParameterEstimator(
            # shift forward so only past returns are used
            np.sqrt(
                returns.ewm(halflife=self.half_life).var().shift(1),
            )
        )
        # initialize cvxpy Parameter(s)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)


class FactorModelRisk(BaseRiskModel):
    """Factor model covariance.

    Args:
        exposures (pd.DataFrame): constant factor exposure matrix or a dataframe
            where the first index is time.
        idyosync (pd.DataFrame or pd.Series): idyosyncratic variances for the symbol,
            either fixed (pd.Series) or through time (pd.DataFrame).
        factor_Sigma (pd.DataFrame or None): a constant factor covariance matrix
            or a DataFrame with multiindex where the first index is time. If None,
            the default, it is understood that the factor covariance is the identity.
            (Otherwise we compute its matrix square root at each step internally and
             apply it to the exposures).
        forecast_error_kappa (float or pd.Series): uncertainty on the
            assets' correlations. See the paper, pages 32-33.

    """

    factor_Sigma = None

    def __init__(
            self,
            exposures,
            idyosync,
            factor_Sigma=None,
            forecast_error_kappa=0.0,
            **kwargs):
        self.exposures = ParameterEstimator(exposures)
        if not (factor_Sigma is None):
            self.factor_Sigma = ParameterEstimator(factor_Sigma)
        self.idyosync = ParameterEstimator(idyosync)
        # ParameterEstimator(forecast_error_kappa)
        self.forecast_error_kappa = forecast_error_kappa
        if ((np.isscalar(forecast_error_kappa) and forecast_error_kappa > 0)
                    or (np.any(forecast_error_kappa > 0))
                ) and factor_Sigma is not None:
            raise NotImplementedError(
                "You should do a Cholesky decomposition of the factor_Sigmas and apply them to the exposures."
            )

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)
        self.idyosync_sqrt = cvx.Parameter(returns.shape[1])
        if not (self.factor_Sigma is None):
            self.factor_Sigma_sqrt = cvx.Parameter(self.factor_Sigma.shape)
        self.forecast_error_penalizer = cvx.Parameter(
            returns.shape[1], nonneg=True)

    def values_in_time(
            self,
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs):
        super().values_in_time(
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs)
        self.idyosync_sqrt.value = np.sqrt(self.idyosync.value)
        if not (self.factor_Sigma is None):
            self.factor_Sigma_sqrt.value = scipy.linalg.sqrtm(
                self.factor_Sigma.value)
            assert np.allclose(
                self.factor_Sigma.value,
                self.factor_Sigma_sqrt.value @ self.factor_Sigma_sqrt.value.T)
        self.forecast_error_penalizer.value = np.sqrt(
            np.sum(self.exposures.value**2, axis=0) + self.idyosync.value)

    def compile_to_cvxpy(self, w_plus, z, value):
        self.expression = cvx.sum_squares(
            cvx.multiply(
                self.idyosync_sqrt,
                (w_plus - self.benchmark_weights)))
        assert self.expression.is_dcp(dpp=True)
        if not (self.factor_Sigma is None):
            self.expression += cvx.sum_squares(
                self.factor_Sigma_sqrt.T @ (w_plus - self.benchmark_weights))
            # self.expression += cvx.quad_form((w_plus.T @ self.exposures.T).T, self.factor_Sigma)
            assert self.expression.is_dcp(dpp=True)
        else:
            self.expression += cvx.sum_squares(self.exposures @
                                               (w_plus - self.benchmark_weights))
            assert self.expression.is_dcp(dpp=True)

        # forecast error risk, assuming factor_Sigma is the identity
        self.expression += self.forecast_error_kappa * cvx.square(
            cvx.abs(
                w_plus -
                self.benchmark_weights).T @ self.forecast_error_penalizer
            # @ cvx.sqrt(cvx.sum(cvx.square(self.exposures), axis=0) + self.idyosync)
        )
        assert self.expression.is_dcp(dpp=True)
        return self.expression


class LowRankRollingRisk(BaseRiskModel):
    """Use `lookback` past returns to build low-rank structured risk model.

    Use this rather that RollingWindowFullCovariance when there are (many) more
     assets than lookback periods, to limit overfit and avoid allocating large matrices.
    (Or, better, use a factor model covariance).

    Args:
        lookback (int): how many past returns are used at each point in time.
            Default 250.
    """

    def __init__(self, lookback=250, zero_cash_risk=True):
        self.lookback = lookback
        self.zero_cash_risk = zero_cash_risk

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Function to initialize object with full prescience."""
        self.recent_returns = cvx.Parameter(
            shape=(self.lookback, returns.shape[1]))
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)

    def values_in_time(
            self,
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs):
        val = past_returns.iloc[-self.lookback:].copy(deep=True)
        if self.zero_cash_risk:
            val.iloc[:, -1] = 0.0
        self.recent_returns.value = val.values
        # update attributes
        super().values_in_time(
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs)

    def compile_to_cvxpy(self, w_plus, z, value):
        self.expression = (cvx.sum_squares(
            self.recent_returns @ (w_plus - self.benchmark_weights)) / self.lookback)
        assert self.expression.is_dcp(dpp=True)
        return self.expression


class RollingWindowFactorModelRisk(FactorModelRisk):
    """Build FactorModelRisk model automatically with rolling window and svd.

    Procedure is detailed in the paper, pages 59-60, although here it is simplified.

    Eventually this will also have logic to store the computed exposures
    and idyosyncratic variances.

    Args:
        lookback (int): how many past returns are used at each point in time
            to estimate the risk model. Default 250.
        num_factors (int): how many factors in the risk model. Default is 1.
        zero_cash_risk (bool): whether to set the column and row of the (implied)
            resulting covariance matrix to zero. Default True.
    """

    def __init__(
            self,
            lookback=250,
            num_factors=1,
            zero_cash_risk=True,
            forecast_error_kappa=0.0):
        self.lookback = lookback
        self.num_factors = num_factors
        self.zero_cash_risk = zero_cash_risk
        self.forecast_error_kappa = forecast_error_kappa

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Function to initialize object with full prescience."""

        self.idyosync = cvx.Parameter(returns.shape[1])
        self.exposures = cvx.Parameter((self.num_factors, returns.shape[1]))
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)

    def values_in_time(
            self,
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs):
        val = past_returns.iloc[-self.lookback:].copy(deep=True)
        if self.zero_cash_risk:
            val.iloc[:, -1] = 0.0

        total_variances = np.sum(val**2, axis=0) / self.lookback
        u, s, v = np.linalg.svd(val, full_matrices=False)

        self.exposures.value = (
            v[: self.num_factors].T * s[: self.num_factors]).T / np.sqrt(self.lookback)

        self.idyosync.value = (
            total_variances -
            np.sum(
                self.exposures.value**2,
                axis=0)).values
        assert np.all(self.idyosync.value >= 0.0)

        super().values_in_time(
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs)


class WorstCaseRisk(BaseRiskModel):
    """Select the most restrictive risk model for each value of the allocation vector.

    Given a list of risk models, penalize the portfolio allocation by the
    one with highest risk value at the solution point. If uncertain about
    which risk model to use this procedure can be an easy solution.

        Args:
            riskmodels (BaseRiskModel): list of BaseRiskModel classes. If using
                non-cash benchmarks, they should be set to each risk model
                individually. Calling set_benchmark on this class has no effect.
    """

    def __init__(self, riskmodels):
        self.riskmodels = riskmodels

    def pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
        """Initialize objects."""
        for risk in self.riskmodels:
            risk.pre_evaluation(
                returns,
                volumes,
                start_time,
                end_time,
                **kwargs)
        super().pre_evaluation(returns, volumes, start_time, end_time, **kwargs)

    def values_in_time(
            self,
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs):
        """Update parameters."""
        for risk in self.riskmodels:
            risk.values_in_time(
                t,
                current_weights,
                current_portfolio_value,
                past_returns,
                past_volumes,
                **kwargs)
        super().values_in_time(
            t,
            current_weights,
            current_portfolio_value,
            past_returns,
            past_volumes,
            **kwargs)

    def compile_to_cvxpy(self, w_plus, z, value):
        risks = [risk.compile_to_cvxpy(w_plus, z, value)
                 for risk in self.riskmodels]
        return cvx.max(cvx.hstack(risks))



# Copyright 2016-2020 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.
# Copyright 2023- The Cvxportfolio Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the MarketSimulator class, which strives to
simulate as accurately as possibly what would have been the realized
performance of a trading policy if it had been run in the market in the past.
In financial jargon this is called *backtesting*.
"""

import copy
import logging
import time
from pathlib import Path


import multiprocess
import numpy as np
import pandas as pd
import cvxpy as cvx

from .result import SimulationResult
from .costs import BaseCost
from .data import FredRate, Yfinance, TimeSeries
from .returns import MultipleReturnsForecasts, ReturnsForecast
from .estimator import Estimator, DataEstimator


class MarketSimulator(Estimator):
    """This class implements a simulator of market performance for trading strategies.

    We strive to make the parameters here as accurate as possible. The following is
    accurate as of 2023 using numbers obtained on the public website of a
    [large US-based broker](https://www.interactivebrokers.com/).


    Args:

        universe (list): list of [Yahoo Finance](https://finance.yahoo.com/) tickers on which to
            simulate performance of the trading strategy. If left unspecified you should at least
            pass `returns` and `volumes`. If you define a different market data access interface
            (look in cvxportfolio.data for how to do it) you should pass instead
            the symbol names for that data provider. Default is empty list.

        returns (pandas.DataFrame): historical open-to-open returns. Default is None, it is ignored
            if universe is specified.

        volumes (pandas.DataFrame): historical market volumes expressed in value (e.g., US dollars).
            Default is None, it is ignored if universe is specified.

        prices (pandas.DataFrame): historical open prices. Default is None, it is ignored
            if universe is specified. These are used to round the trades to integer number of stocks
            if round_trades is True, and compute per-share transaction costs (if `per_share_fixed_cost`
            is greater than zero).

        spreads (pandas.DataFrame): historical bid-ask spreads expressed as (ask-bid)/bid. Default is None,
            equivalent to 0.0. Practical spreads are negligible on US liquid stocks.

        round_trades (bool): round the trade weights provided by a policy so they correspond to an integer
            number of stocks traded. Default is True using Yahoo Finance open prices.

        per_share_fixed_cost (float): transaction cost per share traded. Default value is 0.005 (USD), uses
             Yahoo Finance open prices to simulate the number of stocks traded. See
            https://www.interactivebrokers.com/en/pricing/commissions-home.php

        transaction_cost_coefficient_b (float, pd.Series, or pd.DataFrame): coefficient that multiplies the non-linear
            term of the transaction cost. Default value is 1, you can pass any other constant value, a per-stock Series,
            or a per-day and per-stock DataFrame

        transaction_cost_exponent (float): exponent of the non-linear term of the transaction cost model. Default value 1.5,
             this is applied to the trade volume (in US dollars) over the total market volume (in US dollars). See the
            paper for more details; this model is supported by a long tradition of research in market microstructure.

        rolling_window_sigma_estimator (int): we use an historical rolling standard deviation to estimate the average
            size of the return on a stock on each day, and this multiplies the second term of the transaction cost model.
             See the paper for an explanation of the model. Here you specify the length of the rolling window to use,
             default is 1000.

        spread_on_borrowing_stocks_percent (float, pd.Series, pd.DataFrame): when shorting a stock,
            you will pay a rate on the value
            of the position equal to the cash return plus this spread, expressed in percent annualized. These
            values are hard to find historically, if you are unsure consider long-only portfolios or look
            at CFDs/futures instead. We set the default value to 0.5 (percent annualized) which is probably OK
            for US large cap stocks. See https://www.interactivebrokers.com/en/pricing/short-sale-cost.php

        spread_on_long_positions_percent (float, None, pd.Series, pd.DataFrame): if you trade CFDs you will pay interest
            on your long positions
            as well as your short positions, equal to the cash return plus this value (percent annualized). If
             instead this is None, the default value, you pay nothing on your long positions (as you do if you trade
            stocks). We don't consider dividend payments because those are already incorporated in the
            open-to-open returns as we compute them from the Yahoo Finance data. See cvxportfolio.data for details.

        dividends (float, pd.DataFrame): if not included in the returns (as they are by the default data interface,
            based on `yfinance`), you can pass a DataFrame of dividend payments which will be credited to the cash
            account (or debited, if short) at each round. Default is 0., corresponding to no dividends.

        spread_on_lending_cash_percent (float, pd.Series): the cash account will generate annualized
            return equal to the cash return minus this number, expressed in percent annualized, or zero if
            the spread is larger than the cash return. For example with USDOLLAR cash,
            if the FRED-DFF annualized rate is 4.8% and spread_on_lending_cash_percent is 0.5
            (the default value), then the uninvested cash in the portfolio generates annualized
            return of 4.3%. See https://www.interactivebrokers.com/en/accounts/fees/pricing-interest-rates.php

        spread_on_borrowing_cash_percent (float, pd.Series): if we instead borrow cash we pay the
            cash rate plus this spread, expressed in percent annualized. Default value is 0.5.
            See https://www.interactivebrokers.com/en/trading/margin-rates.php

        cash_key (str or None): name of the cash account. Default is 'USDOLLAR', which gets downloaded by `cvxportfolio.data`
            as the Federal Funds effective rate from FRED. If None, you must pass the cash returns
            along with the stock returns as its last column.

        base_location (pathlib.Path): base location for (by default, sqlite) storage of data.
            Default is `Path.home() / "cvxportfolio"`. Unused if passing `returns` and `volumes`.
    """

    periods_per_year = 252

    def __init__(
            self,
            universe=[],
            returns=None,
            volumes=None,
            costs=None,
            prices=None,
            spreads=0.,
            round_trades=True,
            per_share_fixed_cost=0.005,
            transaction_cost_coefficient_b=1.,
            transaction_cost_exponent=1.5,
            rolling_window_sigma_estimator=1000,
            spread_on_borrowing_stocks_percent=.5,
            spread_on_long_positions_percent=None,
            dividends=0.,
            spread_on_lending_cash_percent=.5,
            spread_on_borrowing_cash_percent=.5,
            cash_key="USDOLLAR",
            base_location=Path.home() /
            "cvxportfolio"):
        """Initialize the Simulator and download data if necessary."""
        if not len(universe):
            if costs is None: # we allow old simulator syntax for the time being
                if ((returns is None) or (volumes is None)):
                    raise SyntaxError(
                        "If you don't specify a universe you should pass `returns` and `volumes`.")
                if not returns.shape[1] == volumes.shape[1] + 1:
                    raise SyntaxError(
                        "In `returns` you must include the cash returns as the last column (and not in `volumes`).")
            self.returns = DataEstimator(returns)
            self.volumes = DataEstimator(volumes) if not volumes is None else volumes
            self.cash_key = returns.columns[-1]
            self.costs = costs
            if not self.costs is None:
                for cost in self.costs:
                    assert isinstance(cost, BaseCost)
            self.prices = DataEstimator(prices) if prices is not None else None
            if prices is None and self.costs is None:
                if per_share_fixed_cost > 0:
                    raise SyntaxError(
                        "If you don't specify prices you can't request `per_share_fixed_cost` transaction costs.")
                if round_trades:
                    raise SyntaxError(
                        "If you don't specify prices you can't request `round_trades`.")
        else:
            self.universe = universe
            self.cash_key = cash_key
            self.base_location = base_location
            self.prepare_data()

        self.spreads = DataEstimator(spreads)
        self.dividends = DataEstimator(dividends)
        self.round_trades = round_trades
        self.per_share_fixed_cost = per_share_fixed_cost
        self.transaction_cost_coefficient_b = transaction_cost_coefficient_b
        self.transaction_cost_exponent = transaction_cost_exponent
        self.rolling_window_sigma_estimator = rolling_window_sigma_estimator
        self.spread_on_borrowing_stocks_percent = spread_on_borrowing_stocks_percent
        self.spread_on_long_positions_percent = spread_on_long_positions_percent
        self.spread_on_lending_cash_percent = spread_on_lending_cash_percent
        self.spread_on_borrowing_cash_percent = spread_on_borrowing_cash_percent

        # compute my DataEstimator(s)
        self.sigma_estimate = DataEstimator(
            self.returns.data.iloc[:, :-1].rolling(self.rolling_window_sigma_estimator).std().shift(1))

    def prepare_data(self):
        """Build data from data storage and download interfaces.

        This is a first cut, it's doing a for loop when we could instead parallelize
        at the `yfinance` level and use the estimator logic of TimeSeries directly.
        """
        self.database_accesses = {}
        for stock in self.universe:
            print('Updating data...')
            self.database_accesses[stock] = TimeSeries(
                stock, base_location=self.base_location)
            self.database_accesses[stock].pre_evaluation()
        if not self.cash_key == 'USDOLLAR':
            raise NotImplementedError(
                'Currently the only data pipeline built is for USDOLLAR cash')
        self.database_accesses[self.cash_key] = TimeSeries(
            'DFF', source='fred', base_location=self.base_location)
        self.database_accesses[self.cash_key].pre_evaluation()

        # build returns
        self.returns = pd.DataFrame(
            {stock: self.database_accesses[stock].data['Return'] for stock in self.universe})
        self.returns[self.cash_key] = self.database_accesses[self.cash_key].data
        self.returns[self.cash_key] = self.returns[self.cash_key].fillna(
            method='ffill')
        

        # build volumes
        self.volumes = pd.DataFrame(
            {stock: self.database_accesses[stock].data['ValueVolume'] for stock in self.universe})
        

        # build prices
        self.prices = pd.DataFrame(
            {stock: self.database_accesses[stock].data['Open'] for stock in self.universe})
        
        
        # yfinance has some issues with most recent data; we patch it here but this
        # logic should go in .data
        if self.prices.iloc[-5:].isnull().any().any():
            drop_at = self.prices.iloc[-5:].isnull().any(axis=1).idxmax()
            self.prices = self.prices.loc[self.prices.index<drop_at]
            self.returns = self.returns.loc[self.returns.index<drop_at]
            self.volumes = self.volumes.loc[self.volumes.index<drop_at]
        
        # for consistency we must also nan-out the last row of returns and volumes
        self.returns.iloc[-1] = np.nan
        self.volumes.iloc[-1] = np.nan
        
        self.returns = DataEstimator(self.returns)
        self.volumes = DataEstimator(self.volumes)
        self.prices = DataEstimator(self.prices)
        
        
    def round_trade_vector(self, u):
        """Round dollar trade vector u."""
        result = pd.Series(u, copy=True)
        result[:-1] = np.round(u[:-1] / self.prices.current_value) * self.prices.current_value
        result[-1] = -sum(result[:-1])
        return result


    def transaction_costs(self, u):
        """Compute transaction costs at time t for dollar trade vector u.

        Returns a non-positive float.

        Args:
            u (pd.Series): dollar trade vector for all stocks including cash (but the cash
                term is not used here).
        """

        result = 0.
        if self.prices is not None:
            # compute number of shares traded.
            # we assume round_trades is True and we add a small number to ensure
            # we are on the safe side of rounding errors
            result += self.per_share_fixed_cost * int(sum(np.abs(u[:-1] + 1E-6) / self.prices.current_value))

        if self.spreads is not None:
            result += sum(self.spreads.current_value * np.abs(u[:-1]))/2.

        result += (np.abs(u[:-1])**self.transaction_cost_exponent) @ (self.transaction_cost_coefficient_b * self.sigma_estimate.current_value / (
            self.volumes.current_value ** (self.transaction_cost_exponent - 1)))
            
        return -result

    def stocks_holding_costs(self, h_plus):
        """Compute holding costs at current time for post trade holdings h_plus (only stocks).

        Args:
            h_plus (pd.Series): post trade holdings vector for all stocks including cash (but the cash
                term is not used here).
        """

        result = 0.
        cash_return = self.returns.current_value[-1]

        # shorting stocks.
        borrowed_stock_positions = np.minimum(h_plus[:-1], 0.)
        result += np.sum((cash_return +
                          (self.spread_on_borrowing_stocks_percent / 100) / self.periods_per_year) *
                         borrowed_stock_positions)

        # going long on stocks.
        if self.spread_on_long_positions_percent is not None:
            long_positions = np.maximum(h_plus[:-1], 0.)
            result -= np.sum((cash_return +
                              (self.spread_on_long_positions_percent /100) / self.periods_per_year) *
                             long_positions)

        # dividends
        result += np.sum(h_plus[:-1] * self.dividends.current_value)

        return result

    def cash_holding_cost(self, h_plus):
        """Compute holding cost on cash (including cash return) for post trade holdings h_plus."""

        cash_return = self.returns.current_value[-1]

        # we subtract from cash the value of borrowed stocks
        # if trading CFDs/futures we must amend this
        real_cash = h_plus[-1] + sum(np.minimum(h_plus[:-1], 0.))

        if real_cash > 0:
            return real_cash * \
                max(cash_return - (self.spread_on_lending_cash_percent/100)/self.periods_per_year, 0.)
        else:
            return real_cash * \
                (cash_return + (self.spread_on_borrowing_cash_percent/100)/self.periods_per_year)
                

    def simulate(self, t, h, policy, **kwargs):
        """Get next portfolio and statistics used by Backtest for reporting.

        The signature of this method differs from other estimators
        because we pass the policy directly to it, and the past returns and past volumes
        are computed by it.
        """
        
        # translate to weights
        current_portfolio_value = sum(h)
        current_weights = h / current_portfolio_value

        # get view of past data
        past_returns = self.returns.data.loc[self.returns.data.index < t]
        past_volumes = self.volumes.data.loc[self.volumes.data.index < t]

        # update all internal estimators
        super().values_in_time(t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs)

        # evaluate the policy
        z = policy.values_in_time(t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs)
        # for safety
        z[-1] = -sum(z[:-1])

        # trades in dollars
        u = z * current_portfolio_value

        # zero out trades on stock that weren't trading on that day
        current_volumes = self.volumes.current_value
        non_tradable_stocks = current_volumes[current_volumes <= 0]
        u[non_tradable_stocks] = 0.

        # round trades
        if self.round_trades:
            u = self.round_trade_vector(u)

        # compute post-trade holdings (including cash balance)
        h_plus = h + u

        # we have updated the internal estimators and they are used by these
        # methods
        transaction_costs = self.transaction_costs(u)
        holding_costs = self.stocks_holding_costs(h_plus)
        cash_holding_costs = self.cash_holding_cost(h_plus)

        # multiply positions by market returns
        h_next = pd.Series(h_plus, copy=True)
        h_next[:-1] *= (1 + self.returns.current_value[:-1])
        
        # credit costs to cash (includes cash return)
        h_next[-1] = h_plus[-1] + (transaction_costs + holding_costs + cash_holding_costs)
            
        return h_next, z, u, transaction_costs, holding_costs, cash_holding_costs
        
    def initialize_policy(self, policy, start_time, end_time):
        """Initialize the policy object.
        
        This method differs from other Estimators because it is the Simulator
        that initializes the policy and all its dependents. It is called by Backtest.
        """
        policy.pre_evaluation(self.returns.data.loc[self.returns.data.index<end_time], 
                              self.volumes.data.loc[self.volumes.data.index<end_time], 
                              start_time, end_time)
        
        
    ### THE FOLLOWING METHODS ARE FROM THE ORIGINAL SIMULATOR (PRE-2023)
    ### The main difference is that you pass a list of costs, which implement
    ### a `value_expression` method, and those take care of the cost evaluation
    ### during backtest. These can be used, as they are in the examples, using
    ### `legacy_run_backtest` and `legacy_run_multiple_backtest`. 
    ###
    ### Also, we have two methods that are not part of the book
    ### and are not well-documented, `what_if` and `attribute`. We may remove
    ### all methods below this comment block in the coming months.

    def propagate(self, h, u, t):
        """Propagates the portfolio forward over time period t, given trades u.

        Args:
            h: pandas Series object describing current portfolio
            u: n vector with the stock trades (not cash)
            t: current time

        Returns:
            h_next: portfolio after returns propagation
            u: trades vector with simulated cash balance
        """
        assert u.index.equals(h.index)

        if self.volumes is not None:
            # don't trade if volume is null
            null_trades = self.volumes.data.columns[self.volumes.data.loc[t] == 0]
            if len(null_trades):
                logging.info(
                    "No trade condition for stocks %s on %s" % (null_trades, t)
                )
                u.loc[null_trades] = 0.0

        hplus = h + u
        costs = [cost.value_expr(t, h_plus=hplus, u=u) for cost in self.costs]
        for cost in costs:
            assert not pd.isnull(cost)
            assert not np.isinf(cost)

        u[self.cash_key] = -sum(u[u.index != self.cash_key]) - sum(costs)
        hplus[self.cash_key] = h[self.cash_key] + u[self.cash_key]

        assert hplus.index.sort_values().equals(
            self.returns.data.columns.sort_values()
        )
        h_next = self.returns.data.loc[t] * hplus + hplus

        assert not h_next.isnull().values.any()
        assert not u.isnull().values.any()
        return h_next, u

    def legacy_run_backtest(
            self,
            initial_portfolio,
            start_time,
            end_time,
            policy,
            loglevel=logging.WARNING):
        """Backtest a single policy."""
        logging.basicConfig(level=loglevel)

        results = SimulationResult(
            initial_portfolio=copy.copy(initial_portfolio),
            policy=policy,
            cash_key=self.cash_key,
            simulator=self,
        )
        h = initial_portfolio

        simulation_times = self.returns.data.index[
            (self.returns.data.index >= start_time)
            & (self.returns.data.index <= end_time)
        ]
        logging.info(
            "Backtest started, from %s to %s"
            % (simulation_times[0], simulation_times[-1])
        )

        for t in simulation_times:
            logging.info("Getting trades at time %s" % t)
            start = time.time()
            try:
                u = policy.get_trades(h, t)
            except cvx.SolverError:
                logging.warning(
                    "Solver failed on timestamp %s. Default to no trades." % t
                )
                u = pd.Series(index=h.index, data=0.0)
            end = time.time()
            assert not pd.isnull(u).any()
            results.log_policy(t, end - start)

            logging.info("Propagating portfolio at time %s" % t)
            start = time.time()
            h, u = self.propagate(h, u, t)
            end = time.time()
            assert not h.isnull().values.any()
            results.log_simulation(
                t=t,
                u=u,
                h_next=h,
                risk_free_return=self.returns.data.loc[t, self.cash_key],
                exec_time=end - start,
            )

        logging.info(
            "Backtest ended, from %s to %s"
            % (simulation_times[0], simulation_times[-1])
        )
        return results

    def legacy_run_multiple_backtest(
        self,
        initial_portf,
        start_time,
        end_time,
        policies,
        loglevel=logging.WARNING,
        parallel=True,
    ):
        """Backtest multiple policies."""

        def _legacy_run_backtest(policy):
            return self.legacy_run_backtest(
                initial_portf, start_time, end_time, policy, loglevel=loglevel
            )

        num_workers = min(multiprocess.cpu_count(), len(policies))
        if parallel:
            workers = multiprocess.Pool(num_workers)
            results = workers.map(_legacy_run_backtest, policies)
            workers.close()
            return results
        else:
            return list(map(_legacy_run_backtest, policies))

    def what_if(self, time, results, alt_policies, parallel=True):
        """Run alternative policies starting from given time."""
        # TODO fix
        initial_portf = copy.copy(results.h.loc[time])
        all_times = results.h.index
        alt_results = self.legacy_run_multiple_backtest(
            initial_portf, time, all_times[-1], alt_policies, parallel
        )
        for idx, alt_result in enumerate(alt_results):
            alt_result.h.loc[time] = results.h.loc[time]
            alt_result.h.sort_index(axis=0, inplace=True)
        return alt_results

    @staticmethod
    def reduce_signal_perturb(initial_weights, delta):
        """Compute matrix of perturbed weights given initial weights."""
        perturb_weights_matrix = np.zeros(
            (len(initial_weights), len(initial_weights)))
        for i in range(len(initial_weights)):
            perturb_weights_matrix[i, :] = initial_weights / (
                1 - delta * initial_weights[i]
            )
            perturb_weights_matrix[i, i] = (1 - delta) * initial_weights[i]
        return perturb_weights_matrix

    def attribute(
            self,
            true_results,
            policy,
            selector=None,
            delta=1,
            fit="linear",
            parallel=True):
        """Attributes returns over a period to individual alpha sources.

        Args:
            true_results: observed results.
            policy: the policy that achieved the returns.
                    Alpha model must be a stream.
            selector: A map from SimulationResult to time series.
            delta: the fractional deviation.
            fit: the type of fit to perform.
        Returns:
            A dict of alpha source to return series.
        """
        # Default selector looks at profits.
        if selector is None:

            def selector(result):
                return result.v - sum(result.initial_portfolio)

        alpha_stream = policy.return_forecast
        assert isinstance(alpha_stream, MultipleReturnsForecasts)
        times = true_results.h.index
        weights = alpha_stream.weights
        assert np.sum(weights) == 1
        alpha_sources = alpha_stream.alpha_sources
        num_sources = len(alpha_sources)
        Wmat = self.reduce_signal_perturb(weights, delta)
        perturb_pols = []
        for idx in range(len(alpha_sources)):
            new_pol = copy.copy(policy)
            new_pol.return_forecast = MultipleReturnsForecasts(
                [ReturnsForecast(el.expected_returns.data) for el in alpha_sources],
                #alpha_sources,
                Wmat[idx, :]
            )
            perturb_pols.append(new_pol)
        # Simulate
        p0 = true_results.initial_portfolio
        alt_results = self.legacy_run_multiple_backtest(
            p0, times[0], times[-1], perturb_pols, parallel=parallel
        )
        # Attribute.
        true_arr = selector(true_results).values
        attr_times = selector(true_results).index
        Rmat = np.zeros((num_sources, len(attr_times)))
        for idx, result in enumerate(alt_results):
            Rmat[idx, :] = selector(result).values
        Pmat = cvx.Variable((num_sources, len(attr_times)))
        if fit == "linear":
            prob = cvx.Problem(cvx.Minimize(0), [Wmat @ Pmat == Rmat])
            prob.solve()
        elif fit == "least-squares":
            error = cvx.sum_squares(Wmat @ Pmat - Rmat)
            prob = cvx.Problem(
                cvx.Minimize(error), [
                    Pmat.T @ weights == true_arr])
            prob.solve()
        else:
            raise Exception("Unknown fitting method.")
        # Dict of results.
        wmask = np.tile(weights[:, np.newaxis], (1, len(attr_times))).T
        data = pd.DataFrame(
            columns=[s.name for s in alpha_sources],
            index=attr_times,
            data=Pmat.value.T * wmask,
        )
        data["residual"] = true_arr - \
            np.asarray((weights @ Pmat).value).ravel()
        data["RMS error"] = np.asarray(
            cvx.norm(Wmat @ Pmat - Rmat, 2, axis=0).value
        ).ravel()
        data["RMS error"] /= np.sqrt(num_sources)
        return data
